import numpy as np
import cv2
import json
import pickle
import datetime
from tqdm import tqdm
from copy import deepcopy
from ..BSpline.central_model import CentralModel, fit_central_model
#from flexiblecc.BSpline.central_model import CentralModel, fit_central_model
from . import initialization
#from flexiblecc.BundleAdjustment import initialization
from socket import gethostname
from scipy.sparse import lil_matrix, csr_matrix
from scipy import optimize
from time import time
from math import ceil
from os import makedirs, path

class BundleAdjustment:
    """
    BundleAdjustment class. This should be used after using parametric camera calibration for an initial guess
    """
    def __init__(self, obj_points, rvecs, tvecs, all_corners_2D, cameraMatrix, distCoeffs, image_dimensions, 
    cm_dimensions=None, cm_stepsize=100, cm_border=0, cm_order=2, cm_fit_control_points=True, 
    cm_knot_method='open_uniform', cm_min_basis_value=1e-4, cm_end_divergence=1e-10, cm_threads=1,
    ls_sparsity=True, ls_verbose=2, ls_ftol=1e-8, ls_gtol=1e-8, ls_method='trf', control_points=None
    ):
        """
        Initializes variables and performs an early fitting of b-spline control points, unless p contains 'cm_fit_control_points=False'

        :param obj_points: Vector of vectors of object points found with calibrate_camera_charuco or calibrate_camera_chessboard from calibrate.py
        :type obj_points: ndarray
        :param rvecs: Rotation vectors estimated for each pattern view found with calibrate_camera_charuco or calibrate_camera_chessboard from calibrate.py
        :type rvecs: ndarray
        :param tvecs: Translation vectors estimated for each pattern view found with calibrate_camera_charuco or calibrate_camera_chessboard from calibrate.py
        :type tvecs: ndarray
        :param all_corners_2D: Vector of vectors of the charuco corners found with calibrate_camera_charuco or calibrate_camera_chessboard from calibrate.py
        :type all_corners_2D: ndarray
        :param cameraMatrix: Intrinsic camera matrix
        :type cameraMatrix: ndarray
        :param distCoeffs: Vector of distortion coefficients of 4, 5, 8, 12 or 14 elements
            (k1,k2,p1,p2[,k3[,k4,k5,k6[,s1,s2,s3,s4[,τx,τy]]]])
        :type distCoeffs: ndarray
        :param image_dimensions: Dimensions of the images used (width, height)
        :type image_dimensions: tuple
        :param cm_dimensions: Dimensions of control points area used in CentralModel (width, height)
        :type cm_dimensions: tuple
        :param cm_stepsize: Distance between control points in pixels
        :type cm_stepsize: int
        :param cm_border: Borders for the CentralModel, changes sample area
        :type cm_border: int
        :param cm_order: Order of the interpolation in CentralModel
        :type cm_order: int
        :param cm_fit_control_points: If True performs a fit of the CentralModel to its control points during initialization
        :type cm_fit_control_points: Boolean
        :param cm_knot_method: Method used to generate the knot vector in CentralModel
        :type cm_knot_method: str
        :param cm_min_basis_value: Used for optimization
        :type cm_min_basis_value: float
        :param cm_end_divergence: A small term added to the tails of the knot vector in CentralModel, when knot method 'open_uniform' is used. Allows for sampling at the endpoints of the spline
        :type cm_end_divergence: float
        :param cm_threads: Amount of threads used for sampling the b-spline from CentralModel (Invalid at the moment, should be 1)
        :type cm_threads: int
        :param ls_sparsity: If True uses sparsity matrix for least squares optimization
        :type ls_sparsity: Boolean
        :param ls_verbose: Verbosity of scipy.least_squares
        :type ls_verbose: int
        :param ls_ftol: scipy.least_squares ftol
        :type ls_ftol: float
        :param ls_gtol: scipy.least_squares gtol
        :type ls_gtol: float
        :param ls_method: scipy.least_squares optimization method
        :type ls_method: str
        :param control_points: Optional control points for the CentralModel class
        :type control_points: ndarray
        """
        start_init = time()

        self.image_dimensions = image_dimensions
        if cm_dimensions is None:
            cm_dimensions = image_dimensions
        self.cm_dimensions = cm_dimensions
        self.cm_stepsize = cm_stepsize
        self.cm_border = cm_border
        self.cm_order = cm_order
        self.cm_fit_control_points = cm_fit_control_points
        self.cm_knot_method = cm_knot_method
        self.cm_min_basis_value = cm_min_basis_value
        self.cm_end_divergence = cm_end_divergence
        self.cm_threads = cm_threads
        self.ls_sparsity = ls_sparsity
        self.ls_verbose = ls_verbose
        self.ls_ftol = ls_ftol
        self.ls_gtol = ls_gtol
        self.ls_method = ls_method
        self.os_info = gethostname()

        self.cm_shape = (int(ceil((self.cm_dimensions[0]+self.cm_border*2)/self.cm_stepsize)+1), \
            int(ceil((self.cm_dimensions[1]+self.cm_border*2)/self.cm_stepsize)+1), 3)

        self.obj_points = obj_points

        self.sampled_spline_rays = deepcopy(self.obj_points)

        if len(self.sampled_spline_rays[0].shape) == 3:
            for i, img in enumerate(self.obj_points):
                self.sampled_spline_rays[i] = img.reshape((img.shape[0],) + (3,))  

        #Reshape board for later use in np.dot()
        if len(self.obj_points[0].shape) == 2:
            for i, img in enumerate(self.obj_points):
                self.obj_points[i] = img.reshape((img.shape[0],) + (3,1))  

        self.transformed_corners_3D = deepcopy(self.obj_points)

        self.all_corners_2D = all_corners_2D

        self.rvecs = np.array(rvecs)
        self.tvecs = np.array(tvecs)

        if control_points != None:
            self.cm_init_ctrl_ptns = None
            self.cm_control_points = control_points.reshape((-1,3))
        else:
            self.cm_init_ctrl_ptns = initialization.grid_creation(self.image_dimensions, cameraMatrix, distCoeffs, border=self.cm_border, step=self.cm_stepsize)
            self.cm_init_ctrl_ptns = np.divide(self.cm_init_ctrl_ptns, np.linalg.norm(self.cm_init_ctrl_ptns, axis=1).reshape((self.cm_init_ctrl_ptns.shape[0], 1)))

            self.cm_control_points = None
            if self.cm_fit_control_points:
                cm, _ = fit_central_model(self.cm_init_ctrl_ptns.reshape(self.cm_shape),
                                        image_dimensions=self.image_dimensions,
                                        grid_dimensions=self.cm_dimensions,
                                        order=self.cm_order,
                                        knot_method=self.cm_knot_method,
                                        min_basis_value=self.cm_min_basis_value,
                                        end_divergence=self.cm_end_divergence,
                                        verbose=self.ls_verbose)
                self.cm_init_ctrl_ptns = cm.a.reshape((-1, 3))

        self.duration_init = time() - start_init


    def transform_board_to_cam(self):
        """
        Returns the board points in the camera frame

        :return transformed_corners_3D: Vector of vectors of board points in the camera's frame
        :rtype transformed_corners_3D: ndarray
        """

        for i, r, t in zip(range(len(self.obj_points)), self.rvecs, self.tvecs):

            rotation_matrix = cv2.Rodrigues(r)[0]
            for j, corner in enumerate(self.obj_points[i]):

                #Might be able to use tensordot instead
                self.transformed_corners_3D[i][j] = np.dot(rotation_matrix, corner) + t

                #Normalize each corner point
                self.transformed_corners_3D[i][j] = np.divide(self.transformed_corners_3D[i][j], np.linalg.norm(self.transformed_corners_3D[i][j]))

        return self.transformed_corners_3D


    def calc_residuals_3D(self, ls_params, return_points_3D=False):
        """
        Returns 3D residuals, this is used for least_squares optimization

        :param ls_params: All parameters changed by least squares, flattened array of control points, rvecs and tvecs
        :type ls_params: ndarray
        :param return_points_3D: Set to true to return residuals, sampled_spline_rays and transformed_corners_3D
        :type return_points_3D: Boolean

        :return residuals: Vector of all residuals
        :rtype residuals: ndarray
        :return sampled_spline_rays: Optional: Vector of vectors of sampled points using the CentralModel class
        :rtype sampled_spline_rays: ndarray
        :return transformed_corners_3D: Optional: Vector of vectors of board points in the camera's frame
        :rtype transformed_corners_3D: ndarray
        """

        n_images = self.all_corners_2D.shape[0]
        cm_shape = self.cm_shape

        self.cm_control_points = ls_params[:np.prod(cm_shape)].reshape(cm_shape)
        self.rvecs = ls_params[np.prod(cm_shape):][:n_images * 3].reshape((n_images, 3, 1))
        self.tvecs = ls_params[np.prod(cm_shape) + n_images * 3:].reshape((n_images, 3, 1))

        cm = CentralModel(
            image_dimensions=self.image_dimensions,
            grid_dimensions=self.cm_dimensions,
            control_points=self.cm_control_points,
            order=self.cm_order,
            knot_method=self.cm_knot_method,
            min_basis_value=self.cm_min_basis_value,
            end_divergence=self.cm_end_divergence)

        # Transform chessboard corners to camera frame
        transformed_corners_3D = np.concatenate(self.transform_board_to_cam())

        # Calculate backward projected rays (b_spline)
        for i, image_corners_2D in enumerate(self.all_corners_2D):
            for j, corner_2D in enumerate(image_corners_2D):
                sample = cm.sample(corner_2D[0, 0], corner_2D[0, 1])
                self.sampled_spline_rays[i][j] = sample#cm.sample(corner_2D[0, 0], corner_2D[0, 1])

        sampled_spline_rays = np.concatenate(self.sampled_spline_rays)
        if return_points_3D:
            transformed_corners_3D.ravel() - sampled_spline_rays.ravel(), sampled_spline_rays, transformed_corners_3D
        else:
            return transformed_corners_3D.ravel() - sampled_spline_rays.ravel()


    def calc_residuals_2D(self, ls_params, return_points_2D=False, verbose=0):
        """
        Returns 2D residuals, used to compare with parametric camera calibration

        :param ls_params: All parameters changed by least squares, flattened array of control points, rvecs and tvecs
        :type ls_params: ndarray
        :param return_points_2D: Set to true to return residuals, estimated_corners_2D and corners_2D
        :type return_points_2D: Boolean

        :return residuals: Vector of all residuals
        :rtype residuals: ndarray
        :return estimated_corners_2D: Optional: Vector of vectors of estimated corners in 2D
        :rtype estimated_corners_2D: ndarray
        :return corners_2D: Optional: Vector of vectors of correct corners in 2D
        :rtype corners_2D: ndarray
        """

        cm_shape = self.cm_shape
        image_size = self.image_dimensions
        n_images = self.obj_points.shape[0]

        self.cm_control_points = ls_params[:np.prod(cm_shape)].reshape(cm_shape)
        self.rvecs = ls_params[np.prod(cm_shape):][:n_images * 3].reshape((n_images, 3, 1))
        self.tvecs = ls_params[np.prod(cm_shape) + n_images * 3:].reshape((n_images, 3, 1))

        self.transform_board_to_cam()

        cm = CentralModel(
            image_dimensions=self.image_dimensions,
            grid_dimensions=self.cm_dimensions,
            control_points=self.cm_control_points,
            order=self.cm_order,
            knot_method=self.cm_knot_method,
            min_basis_value=self.cm_min_basis_value,
            end_divergence=self.cm_end_divergence)
            
        # Calculate backward projected rays (b_spline)
        model_points = deepcopy(self.all_corners_2D)
        if verbose == 1:
            iter = tqdm(range(n_images), total=n_images, unit='image')
        else:
            iter = range(n_images)
        for i in iter:
            for j in range(self.obj_points[i].shape[0]):
                model_points[i][j] = cm.forward_sample(self.transformed_corners_3D[i][j,:,0])

        residuals_manhattan = np.concatenate(model_points).ravel() - np.concatenate(self.all_corners_2D).ravel()
        residuals_euclidean = np.linalg.norm(residuals_manhattan.reshape((-1,2)), axis=1)
        if return_points_2D:
            return rms(residuals_euclidean), residuals_euclidean, model_points, self.all_corners_2D
        else:
            return residuals_euclidean

    
    def get_sparsity_matrix(self, cm_shape, image_dimensions):
        """
        Returns sparsity matrix, used to speed up least squares optimization

        :param cm_shape: Shape of control points for CentralModel
        :type ls_params: tuple
        :param image_dimensions: Dimensions of images used
        :type return_points_2D: tuple

        :return A: A sparse list of lists matrix from scipy.sparse.lil_matrix
        :rtype A: List of Lists
        """
        m = 0 # total number of residuals
        for j in range(self.obj_points.shape[0]):
            m += np.prod(self.obj_points[j].shape)
        
        n_spline_params = np.prod(cm_shape)  # number of b-spline parameters
        n_rvecs_params = self.obj_points.shape[0] * 3  # number of rotation vector parameters
        n_tvecs_params = self.obj_points.shape[0] * 3  # number of translation vector parameters
        n = n_spline_params + n_rvecs_params + n_tvecs_params  # total number of parameters
        A = lil_matrix((m, n), dtype=int)  # sparse matrix describing the connection between parameters and residuals

        # Fill out sparsity matrix for the b-spline (central model) parameters.
        spline_grid = np.ndarray((cm_shape))
        cm = CentralModel(image_dimensions, image_dimensions, spline_grid, 2)

        all_corners_2D = np.concatenate(self.all_corners_2D).reshape((-1, 2))
        for i in range(len(all_corners_2D)):
            control_points = cm.active_control_points(all_corners_2D[i, 0], all_corners_2D[i, 1])
            control_points = control_points.reshape((-1, 2))

            temp = np.zeros(cm_shape)
            for point in control_points:
                temp[int(point[0]), int(point[1])] = np.array([1, 1, 1])

            A[i * 3:i * 3 + 3, :n_spline_params] = np.vstack([temp.ravel()] * 3)

        #Fill out sparsity matrix for rvecs and tvecs parameters
        for i in range(n_rvecs_params):
            img_index = int(i / 3)

            prev_residuals = 0
            for j in range(img_index):
                prev_residuals += np.prod(self.obj_points[j].shape)

            residuals_this_img = np.arange(np.prod(self.obj_points[img_index].shape))

            # Fill out sparsity matrix for rvecs
            A[prev_residuals + residuals_this_img, n_spline_params + i] = 1

            # Fill out sparsity matrix for tvecs
            A[prev_residuals + residuals_this_img, n_spline_params + n_rvecs_params + i] = 1

        return A


    def least_squares(self, folder=None): 
        """
        Performs least squares optimization and returns a scipy.optimize.OptimizeResult object

        :param folder: If folder is specified, the results and parameters are printed to a .pickle file and parameters to a .json file in that folder. 
        The files will be named res_<timestamp>.pickle and par_<timestamp>.json
        :type folder: str

        :return CentralModel: BSpline.central_model.CentralModel object after optimization
        :rtype CentralModel: BSpline.central_model.CentralModel
        :return res: scipy.optimize.OptimizeResult object after optimization
        :rtype res: scipy.optimize.OptimizeResult
        :return rvecs: Rotation vectors estimated for each pattern view
        :rtype rvecs: ndarray
        :return tvecs: Translation vectors estimated for each pattern view
        :rtype tvecs: ndarray
        """
        start_fitting = time()

        if self.cm_control_points is None:
            self.cm_control_points = self.cm_init_ctrl_ptns

        ls_params = np.hstack((self.cm_control_points.ravel(), self.rvecs.ravel(), self.tvecs.ravel()))

        A = None #Sparsity matrix, describes the relationship between parameters (cm_control_points, rvecs, tvecs) and residuals.
        if self.ls_sparsity:
            print('Generating sparsity matrix')
            A = self.get_sparsity_matrix(cm_shape=self.cm_shape, image_dimensions=self.image_dimensions)

        print('Performing least squares optimization on the control points and position of the chessboards')

        res = optimize.least_squares(
            fun=self.calc_residuals_3D,
            x0=ls_params,
            jac_sparsity=A,
            verbose=self.ls_verbose,
            x_scale='jac',
            ftol=self.ls_ftol,
            gtol=self.ls_gtol,
            method=self.ls_method
            )
        
        duration_fitting = time() - start_fitting

        n_images = len(self.rvecs)

        self.cm_control_points = res.x[:np.prod(self.cm_shape)].reshape(self.cm_shape)
        self.rvecs = res.x[np.prod(self.cm_shape):][:n_images * 3].reshape((n_images, 3, 1))
        self.tvecs = res.x[np.prod(self.cm_shape) + n_images * 3:].reshape((n_images, 3, 1))

        self.res = res

        if folder is not None:
            self._print_to_files(res, duration_fitting, folder=folder)

        cm = CentralModel(
            image_dimensions=self.image_dimensions,
            grid_dimensions=self.cm_dimensions,
            control_points=self.cm_control_points,
            order=self.cm_order,
            knot_method=self.cm_knot_method,
            min_basis_value=self.cm_min_basis_value,
            end_divergence=self.cm_end_divergence)

        return cm, res, self.rvecs, self.tvecs


    def rms(self, residuals):
        """
        Calculates and returns root mean squared value.

        :param residuals: Vector of residuals
        :type residuals: ndarray

        :return rms: root mean squared value of residuals
        :rtype rms: float

        """
        return np.sqrt(residuals.dot(residuals)/residuals.size)


    def _print_to_files(self, res, duration_fitting, folder='tests/'):
        """
        Prints results and parameters to files in the folder specified in the arguments.

        :param res: scipy.optimize.OptimizeResult object from least_squares
        :type res: scipy.optimize.OptimizeResult
        :param duration_fitting: Time used for least squares optimization
        :type duration_fitting: float
        :param folder: Path to dump results and paramters to
        :type folder: str
        """
        makedirs(folder, exist_ok=True)

        for key in res:
            if isinstance(res[key], csr_matrix):
                res[key] = res[key].toarray()
            if isinstance(res[key], np.ndarray):
                res[key] = res[key].tolist()

        filename = path.join(folder, '{}_' + datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S') + '{}')
        
        p = {
            'image_dimensions': self.image_dimensions,
            'cm_dimensions': self.cm_dimensions,
            'cm_stepsize': self.cm_stepsize,
            'cm_border': self.cm_border,
            'cm_shape': self.cm_shape,
            'cm_order': self.cm_order,
            'cm_fit_control_points': self.cm_fit_control_points,
            'cm_knot_method': self.cm_knot_method,
            'cm_min_basis_value': self.cm_min_basis_value,
            'cm_end_divergence': self.cm_end_divergence,
            'cm_threads': self.cm_threads,
            'ls_sparsity': self.ls_sparsity,
            'ls_verbose': self.ls_verbose,
            'ls_ftol': self.ls_ftol,
            'ls_gtol': self.ls_gtol,
            'ls_method': self.ls_method,
            'os_info': self.os_info,
            'duration_init': self.duration_init,
            'duration_fitting': duration_fitting
        }

        print('Saving json file with filename: ', filename.format('par', '.json'))
        with open(filename.format('par', '.json'), 'w') as json_out:
            json.dump({'parameters': p}, json_out, ensure_ascii=False, indent=4)
        
        print('Saving pickle file with filename: ', filename.format('res', '.pickle'))
        with open(filename.format('res', '.pickle'), 'wb') as pickle_out:
            pickle.dump([res, p], pickle_out)
    

if __name__ == '__main__':

    error, cameraMatrix, distCoeffs, rvecs, tvecs, _, _, _, \
        all_corners_2D, _, _, _, obj_points_test = np.load('cali.npy', allow_pickle=True)
    
    start_time = time()

    ba = BundleAdjustment(obj_points_test, rvecs, tvecs, all_corners_2D, cameraMatrix, distCoeffs, image_dimensions=(4032,3024),
    cm_dimensions=None, cm_stepsize=2000, cm_border=0, cm_order=2, cm_fit_control_points=True, 
    cm_knot_method='open_uniform', cm_min_basis_value=1e-4, cm_end_divergence=1e-10, cm_threads=1,
    ls_sparsity=True, ls_verbose=2, ls_ftol=1, ls_gtol=1, ls_method='trf', control_points=None)

    cm, res, rvecs, tvecs = ba.least_squares(False)

    pass


import numpy as np
import cv2
import time, datetime
import json
import pickle
from matplotlib import pyplot as plt
from flexiblecc.BSpline.central_model import CentralModel, fit_central_model
from flexiblecc.BundleAdjustment import initialization
from scipy.optimize import least_squares
from scipy.sparse import lil_matrix, csr_matrix
from socket import gethostname
from tqdm import tqdm
from sys import argv


def forward_project(n_images, rvecs, tvecs, checkerboard_points):
    # TODO: make global, don't run all the time

    rays = np.ndarray((n_images, checkerboard_points.shape[1], 3, 1))
    for i, r, t in zip(range(n_images), rvecs, tvecs):
        rotation_matrix = cv2.Rodrigues(r)[0]
        for j in range(checkerboard_points.shape[1]):
            rays[i, j] = np.dot(rotation_matrix, checkerboard_points[i,j]) + t

            # Normalize each ray
            rays[i, j] = np.divide(rays[i, j], np.linalg.norm(rays[i, j]))

    return rays


def residuals_3d(parameters, p, n_images, points_2D, image_size, cm_shape, checkerboard_points):
    spline_grid = parameters[:np.prod(cm_shape)].reshape(cm_shape)
    rvecs = parameters[np.prod(cm_shape):][:n_images * 3].reshape((n_images, 3, 1))
    tvecs = parameters[np.prod(cm_shape) + n_images * 3:].reshape((n_images, 3, 1))

    b_spline_object = CentralModel(
        image_dimensions=image_size,
        grid_dimensions=image_size,
        control_points=spline_grid,
        order=p['cm_order'],
        knot_method=p['cm_knot_method'],
        min_basis_value=p['cm_min_basis_value'],
        end_divergence=p['cm_end_divergence'])

    # Calculate forward projected rays
    forward_projected_rays = forward_project(n_images, rvecs, tvecs, checkerboard_points).ravel()

    # Calculate backward projected rays (b_spline)
    model_rays = np.ndarray((n_images, checkerboard_points.shape[1], 3))
    for i, image_points in zip(range(n_images), points_2D):
        for j, point_2D in zip(range(checkerboard_points.shape[1]), image_points):
            model_rays[i, j] = b_spline_object.sample(point_2D[0, 0], point_2D[0, 1])

    model_rays = model_rays.ravel()

    return forward_projected_rays - model_rays

def residuals_2d(parameters, p, n_images, points_2D, image_size, cm_shape, checkerboard_points):
    spline_grid = parameters[:np.prod(cm_shape)].reshape(cm_shape)
    rvecs = parameters[np.prod(cm_shape):][:n_images * 3].reshape((94, 3, 1))
    tvecs = parameters[np.prod(cm_shape) + n_images * 3:].reshape((94, 3, 1))
    forward_projected_rays = forward_project(n_images, rvecs, tvecs, checkerboard_points)

    b_spline_object = CentralModel(
        image_dimensions=image_size,
        grid_dimensions=image_size,
        control_points=spline_grid,
        order=p['cm_order'],
        knot_method=p['cm_knot_method'],
        min_basis_value=p['cm_min_basis_value'],
        end_divergence=p['cm_end_divergence'])
        
    # Calculate backward projected rays (b_spline)
    model_points = np.ndarray((n_images, checkerboard_points.shape[1], 2))
    for i in tqdm(range(n_images), total=n_images):
        for j in range(checkerboard_points.shape[1]):
            model_points[i, j] = b_spline_object.forward_sample(forward_projected_rays[i, j, :, 0])

    return model_points.ravel() - points_2D.ravel()

def get_sparsity_matrix(n_images, n_points, b_spline_shape, points_2D, image_size):
    m = n_images * n_points * 3  # total number of residuals
    n_spline_params = np.prod(b_spline_shape)  # number of b-spline parameters
    n_rvecs_params = n_images * 3  # number of rotation vector parameters
    n_tvecs_params = n_images * 3  # number of translation vector parameters
    n = n_spline_params + n_rvecs_params + n_tvecs_params  # total number of parameters
    A = lil_matrix((m, n), dtype=int)  # sparse matrix describing the connection between parameters and residuals

    residuals_pr_image = np.arange(n_points * 3)

    # Fill out sparsity matrix for b-spline parameters.
    spline_grid = np.ndarray((b_spline_shape))
    b_spline_object = CentralModel(image_size, image_size, spline_grid, 2)

    points_2D = points_2D.reshape((-1, 2))
    for i in range(len(points_2D)):
        control_points = b_spline_object.active_control_points(points_2D[i, 0], points_2D[i, 1])
        control_points = control_points.reshape((-1, 2))

        temp = np.zeros(b_spline_shape)
        for point in control_points:
            temp[int(point[0]), int(point[1])] = np.array([1, 1, 1])

        A[i * 3:i * 3 + 3, :n_spline_params] = np.vstack([temp.ravel()] * 3)

    # Fill out sparsity matrix for rvecs
    for i in range(n_rvecs_params):
        A[n_points * 3 * int(i / 3) + residuals_pr_image, n_spline_params + i] = 1

    # Fill out sparsity matrix for tvecs
    for i in range(n_tvecs_params):
        A[n_points * 3 * int(i / 3) + residuals_pr_image, n_spline_params + n_rvecs_params + i] = 1

    return A

def old_grid_creation():
    step = 400

    border = 0
    image_width = 1600
    image_height = 1200

    x = np.arange(
        start=-border,
        stop=image_width + border + step,
        step=step,
        dtype=float)

    y = np.arange(
        start=-border,
        stop=image_height + border + step,
        step=step,
        dtype=float)

    src = np.transpose(np.meshgrid(x, y)).reshape((-1, 2))

    fx = 2.89099220e+03
    fy = 2.88229115e+03
    cx = 8.24326073e+02
    cy = 5.63283448e+02

    K = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])

    k1, k2, p1, p2, k3, k4, k5, k6, *_ = 2.03288270e+00, 4.52548448e+01, -7.73226736e-03, 1.80779861e-02, -1.47886929e+02, 2.09683329e+00, 4.65342697e+01, -1.51419436e+02, -1.83082177e-02, -1.92297600e-03, 9.86103858e-03, 7.83822797e-03, 6.18311558e-03, 3.67951499e-02

    distCoeffs = np.array([k1, k2, p1, p2, k3, k4, k5, k6])

    R = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    t = np.array([0, 0, 0])

    Rt = np.ndarray((3, 4))
    Rt[:, :3] = R
    Rt[:, 3] = t

    P = np.dot(K, Rt)

    dst = cv2.undistortPoints(
        src=src,
        cameraMatrix=K,
        distCoeffs=distCoeffs,
        P=P)[:, 0, :]

    Kinv = np.linalg.inv(K)

    projected_pts = np.array([np.dot(Kinv, np.array([[pt[0], ], [pt[1], ], [1, ]])) for pt in dst]).reshape((-1, 3))

    return projected_pts

def bundle_adjustment(p):
    start_time = time.time()

    rvecs, tvecs, checkerboard_points, points_2D = dict(np.load(p['ba_initialization_file'])).values()

    checkerboard_points = checkerboard_points.reshape(checkerboard_points.shape[0], checkerboard_points.shape[2], checkerboard_points.shape[3], 1)

    n_images = len(rvecs)

    # Ligger image_shape, cameraMatrix, distCoeffs i ba_initialization_file?
    # start_grid = initialization.grid_creation(image_shape, cameraMatrix, distCoeffs, border=0, step=400)

    # For at koden stadig virker har jeg insat den gamel igen
    start_grid = old_grid_creation()

    start_grid = np.divide(start_grid, np.linalg.norm(start_grid, axis=1).reshape((start_grid.shape[0], 1)))

    if p['cm_fit_control_points']:
        cm, _ = fit_central_model(start_grid.reshape(p['cm_shape']),
                                  image_dimensions=p['image_size'],
                                  grid_dimensions=p['image_size'],
                                  order=p['cm_order'],
                                  knot_method=p['cm_knot_method'],
                                  min_basis_value=p['cm_min_basis_value'],
                                  end_divergence=p['cm_end_divergence'],
                                  verbose=p['ls_verbose'])
        start_grid = cm.a.reshape((-1, 3))

    ba_params = np.hstack((start_grid.ravel(), rvecs.ravel(), tvecs.ravel()))

    residuals_init = residuals_3d(parameters=ba_params,
                                    p=p,
                                    n_images=n_images,
                                    points_2D=points_2D,
                                    image_size=p['image_size'],
                                    cm_shape=p['cm_shape'],
                                    checkerboard_points=checkerboard_points)

    A = None
    if p['ls_sparsity']:
        A = get_sparsity_matrix(n_images, points_2D.shape[1], p['cm_shape'], points_2D, p['image_size'])
        # plt.spy(A, aspect='auto')
        # plt.show()

    res = least_squares(
        fun=residuals_3d,
        x0=ba_params,
        jac_sparsity=A,
        verbose=p['ls_verbose'],
        x_scale='jac',
        ftol=p['ls_ftol'],
        gtol=p['ls_gtol'],
        method=p['ls_method'],
        args=(p,
              n_images,
              points_2D,
              p['image_size'],
              p['cm_shape'],
              checkerboard_points))

    duration = time.time() - start_time

    spline_grid = res.x[:np.prod(p['cm_shape'])].reshape(p['cm_shape'])
    rvecs = res.x[np.prod(p['cm_shape']):][:n_images * 3].reshape((94, 3, 1))
    tvecs = res.x[np.prod(p['cm_shape']) + n_images * 3:].reshape((94, 3, 1))
    res['initial_residuals'] = residuals_init
    res['start_grid'] = start_grid
    res['spline_grid'] = spline_grid
    res['rvecs'] = rvecs
    res['tvecs'] = tvecs

    return res, duration


if __name__ == '__main__':
    parameters = {
        'image_size': (1600, 1200),
        'checkerboard_shape': (12, 12),
        'cm_dimensions': (1600, 1200),
        'cm_shape': (5, 4, 3),
        'cm_order': 3,
        'cm_fit_control_points': True,
        'cm_knot_method': 'open_uniform',
        'cm_min_basis_value': 0.001,
        'cm_end_divergence': 1e-10,
        'cm_threads': 1,
        'ls_sparsity': True,
        'ls_verbose': 2,
        'ls_ftol': 1e-8,
        'ls_gtol': 1e-8,
        'ls_method': 'trf',
        'ba_initialization_file': 'initial_calibration_results.npz',
        'seed': None,
        'os_info': gethostname(),
        'not_default': []
    }

    for arg in argv[1:]:
        key, value = arg.split('=')
        if key in parameters:
            try:
                if parameters[key] != eval(value):
                    parameters['not_default'].append(key)

                parameters[key] = eval(value)
                print("key '{}' set to '{}' of type '{}'".format(key, parameters[key], type(parameters[key])))
            except:
                parameters[key] = value
                print("key '{}' set to '{}' of type '{}'".format(key, parameters[key], type(parameters[key])))
        else:
            print("key '{}' is not in parameters".format(key))

    if parameters['seed'] != None:
        np.random.seed(parameters['seed'])

    res = pickle.load(open('tests/res_2020-05-05_02-14-53.pickle', 'rb'))

    rvecs, tvecs, checkerboard_points, points_2D = dict(np.load(parameters['ba_initialization_file'])).values()

    checkerboard_points = checkerboard_points.reshape(checkerboard_points.shape[0], checkerboard_points.shape[2], checkerboard_points.shape[3], 1)

    image_size = parameters['image_size']

    cm_shape = parameters['cm_shape']

    n_images = len(rvecs)

    start_grid = old_grid_creation()

    spline_grid = res['spline_grid']
    rvecs = res['rvecs']
    tvecs = res['tvecs']

    ba_params = np.hstack((np.array(spline_grid).ravel(), np.array(rvecs).ravel(), np.array(tvecs).ravel()))

    residuals_2D = residuals_2d(ba_params, parameters, n_images, points_2D, image_size, cm_shape, checkerboard_points)

    ssq = np.sum(residuals_2D**2)

    rms = np.sqrt(residuals_2D.dot(residuals_2D)/residuals_2D.size)

    print('ssq: ', ssq)

    print('Least squares for 2D residuals: ', rms)

    '''
    res, duration = bundle_adjustment(parameters)

    for key in res:
        if isinstance(res[key], csr_matrix):
            res[key] = res[key].toarray()
        if isinstance(res[key], np.ndarray):
            res[key] = res[key].tolist()

    filename = 'tests/{}_' + datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S') + '{}'
    with open(filename.format('par', '.json'), 'w') as json_out:
        json.dump({'parameters': parameters, 'duration': duration}, json_out, ensure_ascii=False, indent=4)

    with open(filename.format('res', '.pickle'), 'wb') as pickle_out:
        pickle.dump(res, pickle_out)
    '''


import numpy as np
from collections.abc import Callable
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plot_distort(undistort_func, image_size, step=10, contour_n_levels=None, save_f_string=None):
    """
    Plots the distorsion discrept width distCoeffs
    Arguments:
        undistort_func (function): A function with argument point coordinates, Nx2 floats. It has to return the undistoret points as a Nx2 array with the same order of points
        image_size (tuple): The size of the image from the camera of length 2
        step (int): Grid step size for undistorted calculations over the image
        contour_n_levels (int/array): Determines the number and positions of the contour lines / regions. If an int n, use n data intervals; i.e. draw n+1 contour lines. The level heights are automatically chosen. If array-like, draw contour lines at the specified levels. The values must be in increasing order.
        save_f_string (str): The f string to use when saving
    """
    assert len(image_size) == 2 and isinstance(image_size, tuple), "image_size has to be a tuple of length 2"
    assert isinstance(undistort_func, Callable), "undistortFunc has to be callable"

    # Make a grid of point there spread over the image
    gridrange = (np.arange(image_size[0], step=step), np.arange(image_size[1], step=step))
    points = np.array(np.meshgrid(*gridrange)).transpose().reshape(-1, 2).astype(np.float32)

    # Calculate the undistorted points
    undistort_points = undistort_func(points)

    # Calculate the difference between the points and there undistorted counter part
    diff = points - undistort_points

    # Calculate the euclidean for the difference
    errors = np.linalg.norm(diff, axis=1)

    # Reshape the errors to the image grid size
    X, Y = np.mgrid[0:len(gridrange[0]), 0:len(gridrange[1])]
    errors_grid = errors.reshape(X.shape)

    # Plot the euclidean distance in pixels between distorted and undistorted
    plt.figure(figsize=(20, 10))

    plt.subplot(1, 2, 1)
    ax = plt.gca()
    plt.title("Euclidean distance in pixels between distorted and undistorted")
    im = plt.imshow(errors.reshape([len(g) for g in gridrange]))
    plt.xticks([], [])
    plt.yticks([], [])

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)

    plt.subplot(1, 2, 2)
    plt.title("Euclidean distance in pixels between distorted and undistorted contour")
    CS = plt.contour(Y, X, errors_grid, levels=contour_n_levels)
    plt.clabel(CS, inline=1, fontsize=10)

    plt.xticks([], [])
    plt.yticks([], [])

    asp = np.diff(CS.ax.get_xlim())[0] / np.diff(CS.ax.get_ylim())[0]
    asp /= np.abs(np.diff(im.axes.get_xlim())[0] / np.diff(im.axes.get_ylim())[0])
    CS.ax.set_aspect(asp)

    #plt.gca().invert_xaxis()
    plt.gca().invert_yaxis()

    if save_f_string is not None:
        plt.savefig(save_f_string.format("distance"))
        plt.close()
    else:
        plt.show()

    plt.figure(figsize=(20, 10))

    # Find the angels the distorted points are moved with
    angels = np.array([np.angle(el[0] + el[1] * 1j, deg=True) for el in diff])

    # Reshape the angels to the image grid size
    X, Y = np.mgrid[0:len(gridrange[0]), 0:len(gridrange[1])]
    angels_grid = angels.reshape(X.shape)

    # Plot the angle between distorted and undistorted
    plt.subplot(1, 2, 1)
    ax = plt.gca()
    plt.title("Angle between distorted and undistorted [deg]")
    im = plt.imshow(angels.reshape([len(g) for g in gridrange]), cmap="hsv")
    plt.xticks([], [])
    plt.yticks([], [])

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax)

    plt.subplot(1, 2, 2)
    plt.title("Angle between distorted and undistorted contour [deg]")
    CS = plt.contour(Y, X, angels_grid, levels=contour_n_levels)
    plt.clabel(CS, inline=1, fontsize=10)

    plt.xticks([], [])
    plt.yticks([], [])

    asp = np.diff(CS.ax.get_xlim())[0] / np.diff(CS.ax.get_ylim())[0]
    asp /= np.abs(np.diff(im.axes.get_xlim())[0] / np.diff(im.axes.get_ylim())[0])
    CS.ax.set_aspect(asp)

    plt.gca().invert_yaxis()

    if save_f_string is not None:
        plt.savefig(save_f_string.format("angle"))
        plt.close()
    else:
        plt.show()


import scipy.stats as stats
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import multivariate_normal, pearsonr, spearmanr
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns
import pandas as pd
import itertools


def maha_dist(x, mean, cov):
    """
    calgulates the malahonobis distance

    :param x: The point
    :param mean: The mean vector
    :param cov:  The covarian matrix
    :return: The malahonobis distance
    """
    return np.matmul(np.matmul((np.array(x) - mean).transpose(), np.linalg.inv(cov)), (np.array(x) - mean))


def maha_dist_arr(arr, mean, cov):
    """
    calgulates the malahonobis distance

    :param arr: The array of points
    :param mean: The mean vector
    :param cov:  The covarian matrix
    :return: The malahonobis distance
    """
    return [maha_dist(x, mean, cov) for x in arr]


def mean_cov(arr):
    """
    Calgulates the mean vector and covarian matrix
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :return mean: The mean vector
    :return cov: The covarian matrix
    """
    cov = np.cov(arr.T)
    mean = np.mean(arr, axis=0)
    return mean, cov


def plot_3DMVN(mean, cov, start, end, res):
    """
    Plots a 3D MVN. Only works for 1x2 mean and 2x2 cov
    :param mean: The mean vector
    :param cov: The covarian matrix
    :param start: The sample space start
    :param end: The sample space end
    :param res: The sample space resulusion
    :return: The axis for the plot
    """
    assert len(mean) == 2, "plot_3DMVN can only handle p=2 MVN's"

    x = np.linspace(start, end, res)
    y = np.linspace(start, end, res)
    X, Y = np.meshgrid(x, y)
    pos = np.empty(X.shape + (2,))
    pos[:, :, 0] = X
    pos[:, :, 1] = Y

    rv = multivariate_normal(mean, cov)
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.plot_surface(X, Y, rv.pdf(pos), cmap='viridis', linewidth=0)
    return ax


def plot_qqplot(arr, ax=None):
    """
    Makes a QQ-Plot in relations to chi^2
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :return: The axis for the plot
    """
    p = arr.shape[1]
    if ax is None:
        plt.figure()
        ax = plt.gca()

    mean, cov = mean_cov(arr)
    dists = maha_dist_arr(arr, mean, cov)
    (osm, osr), (slope, intercept, r) = stats.probplot(x=dists, dist=stats.chi2(p), plot=ax)
    ax.set_title(
        "qq-plot for $d_j^2$ versus ${\chi}_" + str(arr.shape[1]) + "^2$ distribution, $R^2$=" + "{:.04f}".format(
            r ** 2))
    ax.set_xlabel("Quantiles for $d_j^2$")
    ax.set_ylabel("Quantiles for ${\chi}_" + str(arr.shape[1]) + "^2$ distribution")
    return ax


def plot_box(arr, ax=None, violinplot=False):
    """
    Makes a Box plot
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :return: The axis for the plot
    """
    if ax is None:
        plt.figure()
        ax = plt.gca()
    if not violinplot:
        ax.boxplot(arr)
        ax.set_title("Boxplots")
        ax.set_xlabel("Boxplot for the i'th dimensional variable")
    else:
        ax.violinplot(arr, showmedians=True)
        ax.set_title("Violinplot")
        ax.set_xlabel("Violinplot for the i'th dimensional variable")

    return ax


def plot_seaborn_scatter_matrix(arr, kde=True, names=None):
    """
    Makes a seaborn based scatter matrix
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :param kde: Rather or not to ues kernel density estimation in the diagronal plots
    :param names: A list of the names of the p dimensions. If None will be named "Dimension i"
    :return: The axis for the plot
    """
    axs = sns.pairplot(pd.DataFrame(arr, columns=["Dimension {}".format(i) for i in
                                                  range(1, arr.shape[1] + 1)] if names is None else names),
                       diag_kind="kde" if kde else "hist", kind="reg", markers="+")
    return axs


def plot_matplotlib_scatter_matrix(arr, ax_info=None):
    """
    Makes a matplotlib based scatter matrix
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :param ax_info: [ax, ax_pos], where ax is the axis to plot on and ax_pos is which plot to make (i vs j demisions)
    :return: The axis for the plot
    """
    p = arr.shape[1]

    if ax_info is None:
        fig, axs = plt.subplots(p, p)
        for i in range(0, p):
            for j in range(0, p):
                data_i = arr[:, i]
                data_j = arr[:, j]

                if i == j:
                    axs[i, j].hist(data_i)
                else:
                    axs[i, j].scatter(data_i, data_j)

        return axs
    else:
        ax, ax_pos = ax_info
        i, j = ax_pos
        data_i = arr[:, i]
        data_j = arr[:, j]
        if i == j:
            ax.hist(data_i)
        else:
            ax.scatter(data_i, data_j)
        return ax


def _plot_model_check_in_one_wscatter(arr, violinplot=False):
    """
    Makes a model check plot with a scatter matrix
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :return: The axis for the plot
    """
    p = arr.shape[1]

    rows = p if p % 2 == 0 else p + 1
    cols = p * 2

    box_pos = (0, p)
    box_rowspan = rows // 2
    box_colspan = p

    qq_pos = (box_rowspan, p)
    qq_rowspan = rows // 2
    qq_colspan = p

    fig = plt.figure(constrained_layout=True)
    gs = fig.add_gridspec(rows, cols)
    axs = []

    for i in range(p):
        for j in range(p):
            ax = fig.add_subplot(gs[i, j])
            ax = plot_matplotlib_scatter_matrix(arr, [ax, [i, j]])
            axs.append(ax)

    ax = fig.add_subplot(gs[box_pos[0]:box_pos[0] + box_rowspan, box_pos[1]:box_pos[1] + box_colspan])
    ax = plot_box(arr, ax, violinplot=violinplot)
    axs.append(ax)

    ax = fig.add_subplot(gs[qq_pos[0]:qq_pos[0] + qq_rowspan, qq_pos[1]:qq_pos[1] + qq_colspan])
    ax = plot_qqplot(arr, ax)
    axs.append(ax)

    return axs


def plot_model_check(arr, in_one=False, names=None, violinplot=False, save_f_string=None, **saveargs):
    """
    Makes a model check plot
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :param names: A list of the names of the p dimensions. If None will be named "Dimension i". Only in use when in_one=False
    """
    if in_one:
        _plot_model_check_in_one_wscatter(arr, violinplot=violinplot)
    else:
        plot_seaborn_scatter_matrix(arr, names=names)
        if save_f_string:
            plt.savefig(save_f_string.format("scatter_matrix"), **saveargs)
        plt.show()

        plt.figure()
        ax = plt.gca()
        plot_box(arr, ax, violinplot=violinplot)
        if save_f_string:
            plt.savefig(save_f_string.format("box"), **saveargs)
        plt.show()

        plt.figure()
        ax = plt.gca()
        plot_qqplot(arr, ax)
        if save_f_string:
            plt.savefig(save_f_string.format("qqplot"), **saveargs)
        plt.show()


def one_sample_hotelling_t2(arr, mean_to_test):
    """
    Makes a one-sample hotelling T^2 test
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :param mean_to_test: The mean vector to test for
    :return p_value: The p_value for the test
    """
    n = arr.shape[0]
    p = arr.shape[1]

    mean, cov = mean_cov(arr)
    mean_diff = (mean - mean_to_test)

    temp = np.matmul(mean_diff.T, np.linalg.inv(cov / n))
    test_statistic = np.matmul(temp, mean_diff)

    scale = (n - p) / (p * (n - 1))
    p_value = 1 - stats.f.cdf(scale * test_statistic, p, n - p)  # CDF = P(X <= x) -> 1-CDF = P(X > x)
    return p_value


def zero_mean_test(arr):
    """
    Tests if there is a systematic error. By makeing a one-sample hotelling T^2 test vs a mean vector with zeros
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :return p_value: The p_value for the test
    """
    return one_sample_hotelling_t2(arr, mean_to_test=0.0)


def plot_p_values(p_values, critical_value=None, x_labels=None, ax=None):
    """
    Plots p-values with a stem plot
    :param p_values: The p-values to plot
    :param critical_value: If given a horizontal line is made at the critical_value
    :param x_labels: What the p-values describes. If None 1 to len(p_values) is used
    :return p_value: The p_value for the test
    """
    if ax is None:
        plt.figure()
        ax = plt.gca()

    p_values_x = np.arange(len(p_values))
    y_ticks = list(np.arange(0, 1.1, 0.1))
    container_stem = ax.stem(p_values_x, p_values, use_line_collection=True, bottom=0, label="p-values")

    if critical_value is not None:
        left, right = ax.get_xlim()
        ax.hlines(critical_value, left, right, label="Critical value: {}".format(critical_value))
        ax.set_xlim(left, right)
        y_ticks += [critical_value]

    if x_labels is not None:
        left, right = ax.get_xlim()
        locs, labels = plt.xticks()
        x_labels = [""] + list(x_labels) + [""]
        plt.xticks(ticks=locs, labels=x_labels, rotation=45)
        ax.set_xlim(left, right)

    ax.set_ylabel("p-value")
    ax.legend()
    ax.set_ylim(0, 1.0)
    ax.set_yticks(y_ticks)
    plt.grid(True)
    return ax


def correlation_test(a, b, nonparametric):
    """
    Test for correlation. If nonparametric Spearman correlation test else Pearson correlation test
    :param a: First data arrray (one diamation)
    :param b: Secend data array (one diamation)
    :param nonparametric: Reather or not a nonparametric test has to be used. If not assumption that each dataset is normally distributed is made.
    :return p_value: The p_values
    """
    if nonparametric:
        correlation_coefficient, p_value = spearmanr(a, b)  # nonparametric
    else:
        correlation_coefficient, p_value = pearsonr(a, b)  # parametric: normally distributed
    return correlation_coefficient, p_value


def arr_correlation_test(arr, nonparametric, plot=False, plot_critical_value=None, plot_ax=None):
    """
    Test for correlation. If nonparametric Spearman correlation test else Pearson correlation test
    :param arr: The input data in the shape of (n, p) (n the number of samples, p the number of dimensions)
    :param nonparametric: Reather or not a nonparametric test has to be used. If not assumption that each dataset is normally distributed is made.
    :param plot: Wather or not to plot the p-values
    :param critical_value: If given a horizontal line is made at the critical_value
    :param plot_ax: The axis for the plot if None a new plot axis is made
    :return p_value: The p_values
    :return ax: The axis for the plot. Is only returnt if plot=True
    """
    p = arr.shape[1]

    all_combos_idx = list(itertools.combinations(np.arange(p), 2))
    res = {}

    for a_i, b_i in all_combos_idx:
        _, p_value = correlation_test(arr[:, a_i], arr[:, b_i], nonparametric)
        res[(a_i, b_i)] = p_value

    if plot:
        p_values = list(res.values())
        x_labels = list(map(str, list(res.keys())))
        ax = plot_p_values(p_values, plot_critical_value, x_labels, ax=plot_ax)
        ax.set_title("Correlation test")
        ax.set_xlabel("p-value for correlation between (i'th, j'th) dimensions")
        return (res, ax)
    else:
        return (res)



import numpy as np
import cv2


def grid_creation(image_shape, cameraMatrix, distCoeffs, border=0, step=400):
    image_width, image_height = image_shape

    x = np.arange(
        start=-border,
        stop=image_width + border + step,
        step=step,
        dtype=float)

    cx = np.mean(x)
    cw = image_width/2

    x += cw - cx

    y = np.arange(
        start=-border,
        stop=image_height + border + step,
        step=step,
        dtype=float)

    cy = np.mean(y)
    ch = image_height/2

    y += ch - cy

    src = np.transpose(np.meshgrid(x, y)).reshape((-1, 2))

    R = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    t = np.array([0, 0, 0])

    Rt = np.ndarray((3, 4))
    Rt[:, :3] = R
    Rt[:, 3] = t

    P = np.dot(cameraMatrix, Rt)

    dst = cv2.undistortPoints(
        src=src,
        cameraMatrix=cameraMatrix,
        distCoeffs=distCoeffs,
        P=P)[:, 0, :]

    cameraMatrixinv = np.linalg.inv(cameraMatrix)
    projected_pts = np.array([np.dot(cameraMatrixinv, np.array([[pt[0], ], [pt[1], ], [1, ]])) for pt in dst]).reshape(
        (-1, 3))

    return projected_pts


from central_model import CentralModel
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap

import numpy as np

image_dimensions = (10000, 100)
grid_shape =  (10, 10)
control_points = np.zeros(grid_shape + (3,))
knot_method = 'open_uniform'
min_basis_value = 0
end_divergence = 0.01

cmap = get_cmap('rainbow')

for order in range(1, 4):
    figure_name = '{}_{}'.format(knot_method, order)
    plt_title = '{} - Order: {}'.format(knot_method.capitalize(), order)

    plt.figure(figure_name)

    plt.title(plt_title)
    plt.xlabel('x')
    plt.ylabel('Basis function value')

    cm = CentralModel(
        image_dimensions=image_dimensions, 
        grid_dimensions=image_dimensions,
        control_points=control_points,
        order=order,
        knot_method=knot_method,
        min_basis_value=min_basis_value,
        end_divergence=end_divergence
    )

    pts_i = []

    for i in range(grid_shape[0]):
        pts = np.ndarray((image_dimensions[0], 2))
        for u in range(image_dimensions[0]):
            pts[u][0] = u/image_dimensions[0]
            pts[u][1] = cm._B(i, order, cm.th, u/image_dimensions[0])
        
        pts_i.append(pts[:,1])

        c = cmap(i / (grid_shape[0] - 1))
        plt.plot(pts[:,0], pts[:,1], '-', c=c)
    
    pts = np.sum(pts_i, axis=0)
    ones = np.where(pts == 1)
    edges = (np.min(ones) / image_dimensions[0], np.max(ones) / image_dimensions[0])

    t1 = plt.plot(np.array(range(image_dimensions[0]))/image_dimensions[0], pts, '--', c='black')[0]
    plt.ylim((0, 1.01))
    
    t2 = plt.vlines(edges, 0, 1.01, colors='r')
    plt.legend((t1, t2), ('Sum of basis functions', 'Spline edges'))

plt.show()



import numpy as np
import cv2.cv2 as cv2
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d

def Rt_from_rvectvec(rvec, tvec):
    R, *_  = cv2.Rodrigues(rvec)

    Rt = np.zeros(shape=(4,4))
    Rt[:3, :3] = R
    Rt[:-1, 3] = tvec.flatten()
    Rt[-1,-1] = 1
    return Rt

def _do_homogeneous_Rt(p, Rt):
    if p.shape[0] == 3:
        temp = np.ones(4)
        temp[:3] = p
        p = temp
    assert p.shape[0] == 4, "The shape pf p has to be 3 or 4 for each point"

    new_p = np.matmul(Rt,p)

    new_p_uh = new_p[:3]
    new_p_uh = new_p_uh / new_p[-1]

    return new_p_uh

def do_homogeneous_Rt(p, Rt):
    if len(p.shape) == 1:
        new_ps = _do_homogeneous_Rt(p, Rt)
    else:
        new_ps = np.array([_do_homogeneous_Rt(p_el, Rt) for p_el in p])
    return new_ps


def draw_rtvecs(rvecs, tvecs, obj_points_all, draw_points=True, draw_surfaces=False, ax = None):
    if ax is None:
        ax = plt.axes(projection='3d')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')

    n_images = len(obj_points_all)

    transformed_points = []

    ax.scatter([0], [0], [0], marker='x', c='r')

    for i in range(n_images):
        Rt = Rt_from_rvectvec(rvecs[i], tvecs[i])

        transformed_points.append(do_homogeneous_Rt(obj_points_all[i], Rt))
        
        x = transformed_points[i][:,0]
        y = transformed_points[i][:,1]
        z = -transformed_points[i][:,2]

        if draw_points and draw_surfaces:
            pts = ax.scatter(x, y, z)
            c = pts.get_facecolor()[0]
            ax.plot_trisurf(x, y, z, color=c)

        elif draw_points:
            pts = ax.scatter(x, y, z)
        
        elif draw_surfaces:
            ax.plot_trisurf(x, y, z)

    return ax



import numpy as np


# source of equations: https://www.cl.cam.ac.uk/teaching/1999/AGraphHCI/SMAG/node4.html

def uniform(n, order):
    return np.arange(0, n + order + 1) / (n + order)


def open_uniform(n, order, end_divergence=1e-10):
    t = np.zeros((n + order + 1))

    if end_divergence == 0 or order == 1:
        t[-order:] = 1 + end_divergence
    else:
        t[:order] = np.arange(-end_divergence * order, 0, end_divergence)
        t[-order:] = np.arange(1, 1 + (order - 1) * end_divergence, end_divergence) + end_divergence

    t[order:-order] = np.arange(0, n - order + 1) / (n - order)
    return t


if __name__ == '__main__':
    knot_uniform = uniform(5, 2)
    knot_open = open_uniform(5, 2)

    print('uniform:', knot_uniform)
    print('uniform_open:', knot_open)



import numpy as np
from scipy.optimize import least_squares
from tqdm import tqdm
import flexiblecc.BundleAdjustment.to_be_removed_knot_generators as kg
import multiprocessing


# The implementation of this B-spline-based camera model is based on the description in the
# article "Generalized B-spline Camera Model" by Johannes Beck and Christoph Stiller.

# Resources:
# Wikipedia article on b-splines: https://en.wikipedia.org/wiki/B-spline
# Notes of how to construct b-splines: https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/surface/bspline-construct.html

class CentralModel:
    def __init__(self, image_dimensions, grid_dimensions, control_points, order, knot_method='open_uniform',
                 min_basis_value=0.001, end_divergence=1e-10):
        """Initializes a CentralModel object. \n

        Keyword arguments: \n
        image_dimensions: tuple containing (width, height) of image in pixels. \n
        grid_dimensions: tuple containing (width, height) of grid in pixels. \n
        control_points: numpy 3d array containing unit vectors describing camera intrinsics.\n
        order: the order of the interpolation.\n
        knot_method: the method used to generate the knot vector.\n
        min_basis_value: the minimum basis value. Used for optimization.\n
        end_divergence: a small term added to the tails of the knot vector when 'open_uniform' is used. Allows for sampling at the endpoints of the spline.
        """
        assert knot_method in ['open_uniform', 'uniform'], 'knot method should be one of the implemented methods.'
        assert isinstance(image_dimensions, tuple) and len(
            image_dimensions) == 2, 'image_dimensions must be a 2-dimensional tuple.'
        assert isinstance(grid_dimensions, tuple) and len(
            grid_dimensions) == 2, 'grid_dimensions must be a 2-dimensional tuple.'
        assert isinstance(control_points, np.ndarray) and len(control_points.shape) == 3 and control_points.shape[
            -1] == 3, 'control_points must be a 3-dimensional numpy array with a depth of 3.'
        assert control_points.shape[0] > order and control_points.shape[
            1] > order, 'order must be smaller than grid size.'

        self.image_width = image_dimensions[0]
        self.image_height = image_dimensions[1]

        self.grid_width = grid_dimensions[0]
        self.grid_height = grid_dimensions[1]

        self.n = control_points.shape[0]
        self.m = control_points.shape[1]

        self.a = control_points

        self.order = order

        self.B = {}

        if knot_method == 'open_uniform':
            self.th = kg.open_uniform(self.n, order, end_divergence)
            self.tv = kg.open_uniform(self.m, order, end_divergence)

        if knot_method == 'uniform':
            self.th = kg.uniform(self.n, order)
            self.tv = kg.uniform(self.m, order)

        self.min_basis_value = min_basis_value

    def __B__(self, i, k, t, x):
        """Used to calculate the basis function \n

        Keyword arguments: \n
        i: Index for which to sample grid. \n
        k: Basis function order. \n
        t: Knot vector. \n
        x: Pixel coordinate of sample.   
        """
        if (i, k, tuple(t), x) in self.B:
            return self.B[(i, k, tuple(t), x)]

        # Equation 2
        if k == 0:
            if t[i] <= x < t[i + 1]:
                return 1
            else:
                return 0

        # Equation 3
        term1a = x - t[i]
        term1b = t[i + k] - t[i]
        term1c = self.__B__(i, k - 1, t, x)

        # If term1b is zero, the division will be undefined.
        # 'Solution' suggested in https://en.wikipedia.org/wiki/Talk%3AB-spline#Avoiding_division_by_zero
        if term1b == 0:
            term1a = term1b = 1

        term2a = t[i + k + 1] - x
        term2b = t[i + k + 1] - t[i + 1]
        term2c = self.__B__(i + 1, k - 1, t, x)

        # If term2b is zero, the division will be undefined.
        if term2b == 0:
            term2a = term2b = 1

        B = term1a / term1b * term1c + term2a / term2b * term2c

        self.B[(i, k, tuple(t), x)] = B

        return B

    def _normalize(self, coord, grid_size, image_size):
        """Returns normalized coordinates. \n

        Keyword arguments: \n
        coord: Pixel coordinate in one axis.\n
        grid_size: the grid size in pixels in the same axis.\n
        image_size: the image size in pixels in the same axis.
        """
        t = coord + (grid_size - image_size) / 2
        t /= grid_size - 1

        return t

    def sample(self, u, v):
        """Used to sample the b-spline surface. \n

        Keyword arguments: \n
        u: Horizontal pixel coordinate of sample. \n
        v: Vertical pixel coordinate of sample. 
        """
        assert not (isinstance(u, (np.ndarray, list)) or isinstance(v, (np.ndarray, list))) and all(
            np.isreal([u, v])), 'u and v must be numbers.'

        u = self._normalize(u, self.grid_width, self.image_width)
        v = self._normalize(v, self.grid_height, self.image_height)

        Bh = np.array([self.__B__(i, self.order, self.th, u) for i in range(self.n)])
        Bv = np.array([self.__B__(j, self.order, self.tv, v) for j in range(self.m)])

        # Optimization
        vi = np.where(Bh >= self.min_basis_value)[0]
        vj = np.where(Bv >= self.min_basis_value)[0]

        res = np.full((3,), 0.0)
        for i in vi:
            for j in vj:
                aij = self.a[i, j]
                res += np.multiply(aij, Bh[i] * Bv[j])

        return res

    def sample_grid(self):
        """Used to sample the b-spline surface in the input coordinates corresponding to the control points."""
        xs = np.floor((self.grid_width - 1) / (self.n - 1) * np.arange(0, self.n))
        ys = np.floor((self.grid_height - 1) / (self.m - 1) * np.arange(0, self.m))

        pts = np.transpose(np.meshgrid(xs, ys))

        samples = np.ndarray((self.n, self.m, 3))
        for i in range(0, self.n):
            for j in range(0, self.m):
                samples[i, j] = self.sample(pts[i, j, 0], pts[i, j, 1])

        return samples

    def _task(self, i, pts, output):
        output.put(np.array([np.hstack([pt[0], self.sample(pt[1], pt[2])]) for pt in pts]))

    def sample_many(self, pts, threads=1):
        """Used to sample the b-spline surface at multiple points. This method is multi-threaded and will start additional processes even when thread is set to 1. \n

        Keyword arguments: \n
        pts: (n, 2) array of pixel coordinates. \n
        threads: Amount of processes to spawn.
        """

        assert isinstance(pts, np.ndarray) and len(pts.shape) == 2 and pts.shape[
            1] == 2, '\'pts\' should be a numpy array of shape (n,2) containing the points for which to sample the b-spline.'
        assert isinstance(threads,
                          int) and threads >= 1, '\'threads\' should be a positive integer. Got object with type {}.'.format(
            type(threads))

        results = multiprocessing.Queue()

        _pts = np.ndarray((len(pts), 3))
        _pts[:, 0] = range(len(pts))
        _pts[:, 1:] = pts

        split = np.split(_pts, threads)

        processes = [multiprocessing.Process(target=self._task, args=(i, split[threads - i - 1], results)) for i in
                     range(threads)]

        for process in processes:
            process.start()

        results = np.vstack([results.get() for i in range(threads)])
        results = results[results[:, 0].argsort()]

        for process in processes:
            process.join()

        return results[:, :-1]

    def active_control_points(self, u, v):
        """Returns the indeces of the control points that are used to calculate the point with the given pixel coordinates. \n

        Keyword arguments: \n
        u: Horizontal pixel coordinate. \n
        v: Vertical pixel coordinate.
        """
        assert not (isinstance(u, (np.ndarray, list)) or isinstance(v, (np.ndarray, list))) and all(
            np.isreal([u, v])), 'u and v must be numbers.'

        is_even = lambda x: x % 2 == 0

        dx = self._normalize(u, self.grid_width, self.image_width)
        dy = self._normalize(v, self.grid_height, self.image_height)

        px = (self.a.shape[0] - 1) * dx
        py = (self.a.shape[1] - 1) * dy

        between = lambda a, x, b: max(min(x, b), a)

        px = between(np.floor(self.order / 2), px, self.a.shape[0] - 1 - np.floor(self.order / 2))
        py = between(np.floor(self.order / 2), py, self.a.shape[1] - 1 - np.floor(self.order / 2))

        if is_even(self.order):
            x = np.arange(np.ceil(-0.5 * (self.order + 1)), np.ceil(0.5 * self.order) + 1) + np.round(px)
            y = np.arange(np.ceil(-0.5 * (self.order + 1)), np.ceil(0.5 * self.order) + 1) + np.round(py)

        else:
            x = np.arange(np.ceil(-0.5 * self.order), np.ceil(0.5 * self.order) + 1) + np.floor(px)
            y = np.arange(np.ceil(-0.5 * self.order), np.ceil(0.5 * self.order) + 1) + np.floor(py)

        return np.transpose(np.meshgrid(x, y))


def fit_central_model(target_values, image_dimensions, grid_dimensions, order=3, knot_method='open_uniform',
                      end_divergence=0, min_basis_value=1e-3, verbose=0, initial_values=None):
    """Used to fit the control points so the spline interpolates between the values in 'target_values'. Returns the fitted CentralModel and the least squares fitting results. \n

    Keyword arguments: \n
    target_values: The values the solver will try to fit the b-spline surface to.\n
    image_dimensions: see CentralModel's initialization.\n
    grid_dimensions: see CentralModel's initialization.\n
    order: see CentralModel's initialization.\n
    knot_method: see CentralModel's initialization.\n
    end_divergence: see CentralModel's initialization.\n
    min_basis_value: see CentralModel's initialization.\n
    verbose: Changes how much the function prints while running.\n
    initial_values: Optional numpy array with the shape of 'target_values'. Provides the initial control points to the least squares solver.
    """
    if initial_values == None:
        initial_values = target_values

    assert target_values.shape == initial_values.shape, 'target_values and initial_values must have the same shape.'

    shape = target_values.shape

    def fun(params, target_values, grid_shape, image_dimensions, grid_dimensions, order, knot_method, end_divergence,
            min_basis_value):
        grid = np.reshape(params, grid_shape)

        cm = CentralModel(
            image_dimensions=image_dimensions,
            grid_dimensions=grid_dimensions,
            control_points=grid,
            order=order,
            knot_method=knot_method,
            end_divergence=end_divergence,
            min_basis_value=min_basis_value)

        grid_samples = cm.sample_grid().ravel()
        return target_values - grid_samples

    if verbose > 0:
        print('Starting least squares fit.')

    result = least_squares(fun, initial_values.ravel(), verbose=verbose, args=(
    target_values.ravel(), shape, image_dimensions, grid_dimensions, order, knot_method, end_divergence,
    min_basis_value))
    target_values = np.reshape(target_values, (-1, 3))

    ctrl = result['x'].reshape(shape)

    cm = CentralModel(
        image_dimensions=image_dimensions,
        grid_dimensions=grid_dimensions,
        control_points=ctrl,
        order=order,
        knot_method=knot_method,
        end_divergence=end_divergence,
        min_basis_value=min_basis_value)

    return cm, result


if __name__ == '__main__':
    import time

    shape = (500, 500)

    ## Multithreading Test
    if False:
        x = np.arange(shape[0])
        y = np.arange(shape[1])
        pts = np.transpose(np.meshgrid(x, y)).reshape(-1, 2)

        print('Test underway. It might take some time. Please wait...')

        start = time.time()
        cm = CentralModel(shape, shape, np.random.normal(0, 1, (6, 6, 3)), 4)
        _ = cm.sample_many(pts, 16)
        end = time.time()

        print('Threaded version used {:.2f} s. ({} iter/s)'.format(end - start, int(len(pts) / (end - start))))

        start = time.time()
        cm = CentralModel(shape, shape, np.random.normal(0, 1, (6, 6, 3)), 4)
        _ = cm.sample_many(pts, 1)
        end = time.time()

        print('Non-threaded version used {:.2f} s. ({} iter/s)'.format(end - start, int(len(pts) / (end - start))))

    ## Sparsity Test
    if True:
        dim = (200, 200)
        grid = (5, 5, 3)

        ctrl = np.random.normal(0, 1, grid)

        cm = CentralModel(dim, dim, ctrl, 2)

        n = 10

        pts = []

        for i in range(n):
            pts.append(cm.active_control_points(i * (dim[0] / (n - 1)), 100))

    pass



import numpy as np 
import matplotlib.pyplot as plt
import flexiblecc.Metrics.stats as stats
import json
import cv2

calibrate_retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, \
           perViewErrors, charucoCorners_all, charucoIds_all, markerCorners_all, armarkerIds_all, obj_points_all = np.load('cali.npy', allow_pickle=True)


for i,r in enumerate(rvecs):
    rotation_matrix = cv2.Rodrigues(r)
    


with open(r'C:\Users\Jakob\Documents\GitHub\Flexible-Camera-Calibration\flexiblecc\BundleAdjustment\residuals_2D_2020-05-07_13-44-31.json') as f:
    data = json.load(f)

r = 30

data = np.array(data).reshape((-1,2))

data_without_outliers = np.array([[x,y] for x,y in data if -30 < x < 30 and -30 < y < 30])

outlier_indeces = np.array([i for i,[x,y] in enumerate(data) if np.sqrt(x**2+y**2) > r])

#plt.scatter(data[:,0], data[:,1])
#plt.scatter(data_without_outliers[:,0], data_without_outliers[:,1])

stats.plot_model_check(data_without_outliers)

plt.show()


from scipy.spatial import Voronoi
import matplotlib.pyplot as plt
import numpy as np
from collections.abc import Iterable
import cv2

def _voronoi_finite_polygons_2d(vor, radius=None):
    """
    Reconstruct infinite voronoi regions in a 2D diagram to finite
    regions.

    Taken from: https://gist.github.com/pv/8036995 (Author: Pauli Virtanen)

    Parameters
    ----------
    vor : Voronoi
        Input diagram
    radius : float, optional
        Distance to 'points at infinity'.

    Returns
    -------
    regions : list of tuples
        Indices of vertices in each revised Voronoi regions.
    vertices : list of tuples
        Coordinates for revised Voronoi vertices. Same as coordinates
        of input vertices, with 'points at infinity' appended to the
        end.

    """

    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")

    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()

    # Construct a map containing all ridges for a given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue

        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue

            # Compute the missing endpoint of an infinite ridge

            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        # sort region counterclockwise
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        # finish
        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


def _color_angle(values, deg=True):
    if not deg:
        values = np.rad2deg(values)

    if not isinstance(values, Iterable):
        values = [values]

    if len(values) == 1:
        values = [values[0], 360.0, 360.0]
    if len(values) == 2:
        values = [values[0], values[1], 360.0]

    assert len(values) == 3, "The len(values) has to be 1<=len(values)<=3"
    values = np.array(values) / np.array([1.0, 360.0, 360.0])
    H, V, S = values

    Hm = H / 60

    C = V * S
    X = C * (1 - abs(Hm % 2 - 1))

    if 0 <= Hm <= 1:
        ms = [C, X, 0]
    elif 1 < Hm <= 2:
        ms = [X, C, 0]
    elif 2 < Hm <= 3:
        ms = [0, C, X]
    elif 3 < Hm <= 4:
        ms = [0, X, C]
    elif 4 < Hm <= 5:
        ms = [X, 0, C]
    elif 5 < Hm <= 6:
        ms = [C, 0, X]
    else:
        ms = [0, 0, 0]

    Rm, Gm, Bm = ms

    m = V - C

    RGB = np.array([Rm + m, Gm + m, Bm + m])
    RGB = np.round(RGB * 255).astype(int)

    return RGB


def cal_angles_and_mag(image_points, project_points):
    """
    Calgulates the angels and magnitude for the projectet points in relation to the feature points
    :param image_points: The points object points are found in the image
    :param project_points: The re-project points from object points
    :return diff: The difrence for the points as a (nx2) list
    :return angels: The angels for each of the points
    :return mag: The L2 magnitute for each of the points
    """
    img_points_all = []
    angels_all = []
    diff_all = []
    mag_all = []

    for img_ps, rp_ps in zip(image_points, project_points):
        diff = img_ps - rp_ps

        angels = np.array([np.angle(el[0] + el[1] * 1j, deg=True) for el in diff])
        neg_angels = np.where(angels < 0)[0]
        angels[neg_angels] = 360 + angels[neg_angels]

        mag = np.linalg.norm(diff, axis=1)

        diff_all.append(diff)
        angels_all.append(angels)
        mag_all.append(mag)
        real_img_points = img_ps.reshape(-1, 2)
        img_points_all.append(real_img_points)

    diff_all = np.concatenate(diff_all).reshape(-1, 2)
    angels_all = np.concatenate(angels_all).flatten()
    mag_all = np.concatenate(mag_all).flatten()
    img_points_all = np.concatenate(img_points_all).reshape(-1, 2)
    return img_points_all, diff_all, angels_all, mag_all


def projectPoints_and_cal_angles_and_mag(image_points, obj_points, rvecs, tvecs, cameraMatrix, distCoeffs):
    """
    Calgulates the angels and magnitude for the projectet points in relation to the feature points using rvecs, tvecs, cameraMatrix, distCoeffs
    :param image_points: The points object points are found in the image
    :param obj_points: The object points
    :param rvecs (ndarray): Rotation vectors estimated for each pattern view
    :param tvecs (ndarray): Translation vectors estimated for each pattern view
    :param cameraMatrix: 3x3 floating-point camera matrix
            [[fx, 0,  cx],
             [0,  fy, cy],
    :param distCoeffs: vector of distortion coefficients of 4, 5, 8, 12 or 14 elements
            (k1,k2,p1,p2[,k3[,k4,k5,k6[,s1,s2,s3,s4[,τx,τy]]]])
    :return img_points: The image points in a flat list
    :return diff: The difrence for the points as a (nx2) list
    :return angels: The angels for each of the points
    :return mag: The L2 magnitute for each of the points
    """
    img_points_all = []
    project_points = []

    for img_ps, obj_ps, rvec, tvec in zip(image_points, obj_points, rvecs, tvecs):
        real_img_points = img_ps.reshape(-1, 2)

        repor_img_points = cv2.projectPoints(obj_ps, rvec, tvec, cameraMatrix, distCoeffs)[0].reshape(-1, 2)

        project_points.append(repor_img_points)
        img_points_all.append(real_img_points)

    _, diff_all, angels_all, mag_all = cal_angles_and_mag(img_points_all, project_points)
    img_points_all = np.concatenate(img_points_all).reshape(-1, 2)
    return img_points_all, diff_all, angels_all, mag_all


def plot_voronoi(points, points_angle, deg=True, ax=None, xy_lim=True, radius=None, plot_points=False):
    """
    Plots voronoi diagram for the given points
    """
    if ax is None:
        ax = plt.gca()

    if plot_points:
        ax.plot(points[:, 0], points[:, 1], "o")

    vor = Voronoi(points)
    regions, vertices = _voronoi_finite_polygons_2d(vor, radius=radius)

    for i, region in enumerate(regions):
        polygon = vertices[region]
        ax.fill(*zip(*polygon), color=_color_angle(points_angle[i], deg=deg) / 255)

    if xy_lim:
        ax.set_xlim([vor.min_bound[0], vor.max_bound[0]])
        ax.set_ylim([vor.min_bound[1], vor.max_bound[1]])

    return ax



import cv2
import numpy as np
from tqdm import tqdm
from collections.abc import Iterable, Callable
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import os


def calibrate_camera_chessboard(gray_imgs, pattern_size, win_size=(10, 10), zero_zone=(-1, -1), criteria=None,
                     flags=(cv2.CALIB_RATIONAL_MODEL +
                            cv2.CALIB_THIN_PRISM_MODEL +
                            cv2.CALIB_TILTED_MODEL), verbose=0, draw_chessboards=None):
    """
    Will do a normal camera calibration. This will be done by finding the chessboards in the provided grayscale
    images.

    :param gray_imgs: Array there contains the images with chessboards. The images has to be in grayscale colorspace and in a "int" datatype
    :type gray_imgs: iterable
    :param pattern_size: Number of inner corners per a chessboard row and column
    :type pattern_size: tuple
    :param win_size: Half of the side length of the search window
    :type win_size: tuple
    :param zero_zone: Half of the size of the dead region in the middle of the search zone over which the summation in the formula below is not done. It is used sometimes to avoid possible singularities of the autocorrelation matrix. The value of (-1,-1) indicates that there is no such a size.
    :type zero_zone: tuple
    :param criteria: Criteria for termination of the iterative process of corner refinement. That is, the process of corner position refinement stops either after criteria.maxCount iterations or when the corner position moves by less than criteria.epsilon on some iteration. If None the criteria will be set to (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    :type criteria: tuple
    :param flags: The flags used in cv2.calibrateCameraExtended. Default flags CALIB_RATIONAL_MODEL, CALIB_THIN_PRISM_MODEL and CALIB_TILTED_MODEL. Flages tehre can be used:
            CALIB_USE_INTRINSIC_GUESS cameraMatrix contains valid initial values of fx, fy, cx, cy that are optimized further. Otherwise, (cx, cy) is initially set to the image center ( imageSize is used), and focal distances are computed in a least-squares fashion. Note, that if intrinsic parameters are known, there is no need to use this function just to estimate extrinsic parameters. Use solvePnP instead.
            CALIB_FIX_PRINCIPAL_POINT The principal point is not changed during the global optimization. It stays at the center or at a different location specified when CALIB_USE_INTRINSIC_GUESS is set too.
            CALIB_FIX_ASPECT_RATIO The functions considers only fy as a free parameter. The ratio fx/fy stays the same as in the input cameraMatrix . When CALIB_USE_INTRINSIC_GUESS is not set, the actual input values of fx and fy are ignored, only their ratio is computed and used further.
            CALIB_ZERO_TANGENT_DIST Tangential distortion coefficients (p1,p2) are set to zeros and stay zero.
            CALIB_FIX_K1,...,CALIB_FIX_K6 The corresponding radial distortion coefficient is not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
            CALIB_RATIONAL_MODEL Coefficients k4, k5, and k6 are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the rational model and return 8 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_THIN_PRISM_MODEL Coefficients s1, s2, s3 and s4 are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the thin prism model and return 12 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_FIX_S1_S2_S3_S4 The thin prism distortion coefficients are not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
            CALIB_TILTED_MODEL Coefficients tauX and tauY are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the tilted sensor model and return 14 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_FIX_TAUX_TAUY The coefficients of the tilted sensor model are not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
    :type flags: int
    :param verbose:  0, or 1. Verbosity mode. 0 = silent, 1 = progress bar and print out
    :type verbose: int
    :param draw_chessboards: The file path to save draw chessboards images. If None it will not save any
    :type draw_chessboards: str

    :returns:
    - retval: The RMS re-projection error in pixels
    - cameraMatrix: Output 3x3 floating-point camera matrix
            [[fx, 0,  cx],
             [0,  fy, cy],
             [0,  0,  1]]
    - distCoeffs: Output vector of distortion coefficients of 4, 5, 8, 12 or 14 elements
            (k1,k2,p1,p2[,k3[,k4,k5,k6[,s1,s2,s3,s4[,τx,τy]]]])
    - rvecs: Rotation vectors estimated for each pattern view
    - tvecs: Translation vectors estimated for each pattern view
    - stdDeviationsIntrinsics: Standard deviations estimated for intrinsic parameters
    - stdDeviationsExtrinsics: Standard deviations estimated for extrinsic parameters
    - perViewErrors: Vector of the RMS re-projection error estimated for each pattern view
    - objPoint: Vector of vectors of calibration pattern points in the calibration pattern coordinate space
    - imgPoints: Vector of vectors of the projections of calibration pattern points

    :rtype retval: float
    :rtype cameraMatrix: ndarray
    :rtype distCoeffs: ndarray
    :rtype rvecs: ndarray
    :rtype tvecs: ndarray
    :rtype stdDeviationsIntrinsics: ndarray
    :rtype stdDeviationsExtrinsics: ndarray
    :rtype perViewErrors: ndarray
    :rtype objPoints: ndarray
    :rtype imgPoints: ndarray
    """

    assert isinstance(gray_imgs, Iterable), "gray_imgs has to be a iterable there consists of the grayscale images"
    assert len(pattern_size) == 2 and isinstance(pattern_size, tuple), "pattern_size has to be a tuple of length 2"
    assert len(win_size) == 2 and isinstance(win_size, tuple), "win_size has to be a tuple of length 2"
    assert len(zero_zone) == 2 and isinstance(zero_zone, tuple), "zero_zone has to be a tuple of length 2"

    if criteria is None:
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    else:
        assert isinstance(criteria, tuple), "criteria has to be a tuple"

    if draw_chessboards is not None:
        os.makedirs(draw_chessboards, exist_ok=True)

    # Object points for a chessboard
    objp = np.zeros((1, np.product(pattern_size[:2]), 3), np.float32)
    objp[0, :, :2] = np.mgrid[0:pattern_size[0], 0:pattern_size[1]].transpose().reshape(-1, 2)

    # Arrays to hold the object points and corner point for all the chessboards
    obj_points = []
    img_points = []

    if verbose == 1:
        iter = tqdm(gray_imgs, unit="image")
        print("Finding chessboard pattern in the images")
    else:
        iter = gray_imgs

    for i, gray_img in enumerate(iter):
        # Find roof corners in the images
        pattern_was_found, corners = cv2.findChessboardCorners(gray_img, pattern_size)

        # If there was a chessboard in the image
        if pattern_was_found:
            # Add object points for the chessboard
            obj_points.append(objp)

            # Find better sub pix position for the corners in the roof corners neighbourhood
            new_better_corners = cv2.cornerSubPix(gray_img, corners, win_size, zero_zone, criteria)

            if draw_chessboards is not None:
                img_first = cv2.drawChessboardCorners(gray_img.copy(), pattern_size, corners,
                                                       pattern_was_found)
                img_better = cv2.drawChessboardCorners(gray_img.copy(), pattern_size, new_better_corners,
                                                       pattern_was_found)
                cv2.imwrite(os.path.join(draw_chessboards, "{}_first.png".format(i)), img_first)
                cv2.imwrite(os.path.join(draw_chessboards, "{}_better.png".format(i)), img_better)

            # Add the better corners
            img_points.append(new_better_corners)

    if verbose == 1:
        print("Doing camera calibrate")

    # Do the camera calibrations from the object points and corners found in the images

    retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, perViewErrors = cv2.calibrateCameraExtended(
        obj_points, img_points, (gray_imgs[0].shape[1], gray_imgs[0].shape[0]), cameraMatrix=None, distCoeffs=None,
        flags=flags)

    return retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, \
           perViewErrors, np.array(obj_points), np.array(img_points)


def find_Charuco(gray_imgs, dictionary, board, win_size=(10, 10), zero_zone=(-1, -1), criteria=None, verbose=0, draw=None):
    if criteria is None:
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    else:
        assert isinstance(criteria, tuple), "criteria has to be a tuple"

    cv2.aruco.DetectorParameters_create()

    charucoCorners_all = []
    charucoIds_all = []

    markerCorners_all = []
    markerIds_all = []
    obj_points_all = []

    itr = enumerate(gray_imgs)
    if verbose == 1:
        print("Finding charuco features")
        itr = tqdm(itr, total=len(gray_imgs), unit="image")

    for i, img_gray in itr:
        markerCorners, markerIds, rejectedImgPoints = cv2.aruco.detectMarkers(img_gray, dictionary)

        markers_found = len(markerCorners) > 0

        if markers_found:

            markerCorners_all.append(markerCorners)
            markerIds_all.append(markerIds)

            retval, charucoCorners, charucoIds = cv2.aruco.interpolateCornersCharuco(markerCorners, markerIds, img_gray,
                                                                                     board)

            charuco_corners_found = charucoCorners is not None and len(charucoCorners) > 3

            if charuco_corners_found:
                new_better_charucoCorners = cv2.cornerSubPix(img_gray, charucoCorners.copy(), win_size, zero_zone,
                                                             criteria)

                ids = charucoIds.flatten()
                obj_points = board.chessboardCorners[ids]

                charucoCorners_all.append(new_better_charucoCorners)
                charucoIds_all.append(charucoIds)
                obj_points_all.append(obj_points)
                if draw is not None:
                    img_gray_draw = img_gray.copy()
                    cv2.aruco.drawDetectedMarkers(img_gray_draw, markerCorners, markerIds)
                    cv2.aruco.drawDetectedCornersCharuco(img_gray_draw, new_better_charucoCorners, charucoIds)
                    cv2.imwrite(os.path.join(draw, "{}.png".format(i)), img_gray_draw)

    return markerCorners_all, markerIds_all, charucoCorners_all, charucoIds_all, obj_points_all


def calibrate_camera_charuco(gray_imgs, squaresX, squaresY, squareLength, markerLength, dictionary, win_size=(10, 10),
                     zero_zone=(-1, -1), criteria=None,
                     flags=(cv2.CALIB_RATIONAL_MODEL +
                            cv2.CALIB_THIN_PRISM_MODEL +
                            cv2.CALIB_TILTED_MODEL), verbose=0, draw=None):
    """
    Will do a charuco camera calibration. This will be done by finding the charuco boards in the provided grayscale images.
    -> calibrate_retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, perViewErrors, charucoCorners_all, charucoIds_all, markerCorners_all, armarkerIds_all, obj_points_all, board
    :param gray_imgs: Array there contains the images with chessboards. The images has to be in grayscale colorspace and in a "int" datatype
    :type gray_imgs: iterable
    :param squaresX: number of chessboard squares in X direction
    :type squaresX: int
    :param squaresY: number of chessboard squares in Y direction
    :type squaresY: int
    :param squareLength: chessboard square side length (normally in meters)
    :type squareLength: float
    :param markerLength: marker side length (same unit than squareLength)
    :type markerLength: float
    :param dictionary: dictionary of markers indicating the type of markers. Can normally be found with cv2.aruco.getPredefinedDictionary
    :type dictionary: opencv aruco_Dictionary
    :param win_size: Half of the side length of the search window
    :type win_size: tuple
    :param zero_zone: Half of the size of the dead region in the middle of the search zone over which the summation in the formula below is not done. It is used sometimes to avoid possible singularities of the autocorrelation matrix. The value of (-1,-1) indicates that there is no such a size.
    :type zero_zone: tuple
    :param criteria: Criteria for termination of the iterative process of corner refinement. That is, the process of corner position refinement stops either after criteria.maxCount iterations or when the corner position moves by less than criteria.epsilon on some iteration. If None the criteria will be set to (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    :type criteria: tuple
    :param flags: The flags used in cv2.calibrateCameraExtended. Default flags CALIB_RATIONAL_MODEL, CALIB_THIN_PRISM_MODEL and CALIB_TILTED_MODEL. Flages tehre can be used:
            CALIB_USE_INTRINSIC_GUESS cameraMatrix contains valid initial values of fx, fy, cx, cy that are optimized further. Otherwise, (cx, cy) is initially set to the image center ( imageSize is used), and focal distances are computed in a least-squares fashion. Note, that if intrinsic parameters are known, there is no need to use this function just to estimate extrinsic parameters. Use solvePnP instead.
            CALIB_FIX_PRINCIPAL_POINT The principal point is not changed during the global optimization. It stays at the center or at a different location specified when CALIB_USE_INTRINSIC_GUESS is set too.
            CALIB_FIX_ASPECT_RATIO The functions considers only fy as a free parameter. The ratio fx/fy stays the same as in the input cameraMatrix . When CALIB_USE_INTRINSIC_GUESS is not set, the actual input values of fx and fy are ignored, only their ratio is computed and used further.
            CALIB_ZERO_TANGENT_DIST Tangential distortion coefficients (p1,p2) are set to zeros and stay zero.
            CALIB_FIX_K1,...,CALIB_FIX_K6 The corresponding radial distortion coefficient is not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
            CALIB_RATIONAL_MODEL Coefficients k4, k5, and k6 are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the rational model and return 8 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_THIN_PRISM_MODEL Coefficients s1, s2, s3 and s4 are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the thin prism model and return 12 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_FIX_S1_S2_S3_S4 The thin prism distortion coefficients are not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
            CALIB_TILTED_MODEL Coefficients tauX and tauY are enabled. To provide the backward compatibility, this extra flag should be explicitly specified to make the calibration function use the tilted sensor model and return 14 coefficients. If the flag is not set, the function computes and returns only 5 distortion coefficients.
            CALIB_FIX_TAUX_TAUY The coefficients of the tilted sensor model are not changed during the optimization. If CALIB_USE_INTRINSIC_GUESS is set, the coefficient from the supplied distCoeffs matrix is used. Otherwise, it is set to 0.
    :type flags: int
    :param verbose:  0, or 1. Verbosity mode. 0 = silent, 1 = progress bar and print out
    :type verbose: int
    :param draw: The file path to save draw charuco images. If None it will not save any
    :type draw: str

    :returns:
    - retval: The RMS re-projection error in pixels
    - cameraMatrix: Output 3x3 floating-point camera matrix
            [[fx, 0,  cx],
             [0,  fy, cy],
             [0,  0,  1]]
    - distCoeffs: Output vector of distortion coefficients of 4, 5, 8, 12 or 14 elements
            (k1,k2,p1,p2[,k3[,k4,k5,k6[,s1,s2,s3,s4[,τx,τy]]]])
    - rvecs: Rotation vectors estimated for each pattern view
    - tvecs: Translation vectors estimated for each pattern view
    - stdDeviationsIntrinsics: Standard deviations estimated for intrinsic parameters
    - stdDeviationsExtrinsics: Standard deviations estimated for extrinsic parameters
    - perViewErrors: Vector of the RMS re-projection error estimated for each pattern view
    - charucoCorners: Vector of vectors of the charuco corners found in the images
    - charucoIds: Vector of vectors of the charuco corners id's found in the images
    - markerCorners: Vector of vectors of the charuco corners found in the images
    - marmarkerIds: Vector of vectors of the charuco corners id's found in the images
    - obj_points: Vector of vectors of the charuco corners as object points found in the images
    - board: The used charuco board

    :rtype retval: float
    :rtype cameraMatrix: ndarray
    :rtype distCoeffs: ndarray
    :rtype rvecs: ndarray
    :rtype tvecs: ndarray
    :rtype stdDeviationsIntrinsics: ndarray
    :rtype stdDeviationsExtrinsics: ndarray
    :rtype perViewErrors: ndarray
    :rtype charucoCorners: ndarray
    :rtype charucoIds: ndarray
    :rtype markerCorners: ndarray
    :rtype marmarkerIds: ndarrayv
    :rtype obj_points: ndarray
    :rtype board: aruco_CharucoBoard
    """
    if criteria is None:
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    else:
        assert isinstance(criteria, tuple), "criteria has to be a tuple"

    if draw is not None:
        os.makedirs(draw, exist_ok=True)

    board = cv2.aruco.CharucoBoard_create(squaresX, squaresY, squareLength, markerLength, dictionary)

    markerCorners_all, markerIds_all, charucoCorners_all, charucoIds_all, obj_points_all = find_Charuco(gray_imgs,
                                                                                                        dictionary,
                                                                                                        board,
                                                                                                        win_size,
                                                                                                        zero_zone,
                                                                                                        criteria,
                                                                                                        verbose,draw)

    if verbose == 1:
        print("Doing camera calibrate")

    img_shape = gray_imgs[0].shape[:2]

    calibrate_retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, perViewErrors = cv2.aruco.calibrateCameraCharucoExtended(charucoCorners_all, charucoIds_all, board, img_shape, cameraMatrix=None, distCoeffs=None, flags=flags)

    if draw is not None:
        for i, (charucoCorners, charucoIds, rvec, tvec) in enumerate(
                zip(charucoCorners_all, charucoIds_all, rvecs, tvecs)):
            img_f = os.path.join(draw, "{}.png".format(i))
            img = cv2.imread(img_f)
            cv2.aruco.drawAxis(img, cameraMatrix, distCoeffs, rvec, tvec, 0.1)
            cv2.imwrite(img_f, img)

    charucoCorners_all, charucoIds_all = np.array(charucoCorners_all), np.array(charucoIds_all)
    markerCorners_all, armarkerIds_all = np.array(markerCorners_all), np.array(markerIds_all)
    obj_points_all = np.array(obj_points_all)

    if verbose == 1:
        print("Calibration done")

    return calibrate_retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, \
           perViewErrors, charucoCorners_all, charucoIds_all, markerCorners_all, armarkerIds_all, obj_points_all, board


import numpy as np
import pickle
import json
from bundle_adjustment import BundleAdjustment
from matplotlib import pyplot as plt

def rms(residuals):
    return np.sqrt(residuals.dot(residuals)/residuals.size)

res, p = pickle.load(open('tests/res_2020-05-09_22-32-50.pickle', 'rb'))

p['ba_initialization_file'] = 'cali.npy'

error, cameraMatrix, distCoeffs, _, _, _, _, _, \
        all_corners_2D, _, _, _, obj_points = np.load(p['ba_initialization_file'], allow_pickle=True)

n_images = obj_points.shape[0]

res.x = np.array(res.x)

cm_control_points = res.x[:np.prod(p['cm_shape'])].reshape(p['cm_shape'])
rvecs = res.x[np.prod(p['cm_shape']):][:n_images * 3].reshape((n_images, 3, 1))
tvecs = res.x[np.prod(p['cm_shape']) + n_images * 3:].reshape((n_images, 3, 1))

ba = BundleAdjustment(obj_points, rvecs, tvecs, all_corners_2D, cameraMatrix, distCoeffs, cm_control_points)

ls_params = np.hstack((cm_control_points.ravel(), rvecs.ravel(), tvecs.ravel()))

#residuals_3D = ba.calc_residuals_3D(ls_params, p)

residuals_2D, estimated_points_2D, correct_points_2D = ba.calc_residuals_2D(ls_params, return_points_2D=True)

np.save('residuals_and_points_2D_2020-05-09_22-32-50', [residuals_2D, estimated_points_2D, correct_points_2D])

#with open('residuals_2D_2020-05-09_22-32-50.json', 'w') as json_out:
#    json.dump(residuals_2D.tolist(), json_out, ensure_ascii=False, indent=4)

print('rms: ', rms(residuals_2D))


import numpy as np

p = {
    'image_size': (1600, 1200),
    'checkerboard_shape': (12, 12),
    'cm_dimensions': (1600, 1200),
    'cm_shape': (5, 4, 3),
    'cm_order': 3,
    'cm_fit_control_points': True,
    'cm_knot_method': 'open_uniform',
    'cm_min_basis_value': 0.001,
    'cm_end_divergence': 1e-10,
    'cm_threads': 1,
    'ls_sparsity': True,
    'ls_verbose': 2,
    'ls_ftol': 1,
    'ls_gtol': 1e-8,
    'ls_method': 'trf',
    'ba_initialization_file': 'initial_calibration_results.npz',
    'seed': None,
    'not_default': []
}


rvecs, tvecs, checkerboard_points, points_2D = dict(np.load(p['ba_initialization_file'])).values()

count = 0
for checkerboard in checkerboard_points:
    if checkerboard.all() == checkerboard_points[0].all():
        count += 1
print(count)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="flexiblecc",
    version="0.1.14",
    author="Simon L. B. Sørensen and contributors",
    author_email="simonlyckbjaert@hotmail.com",
    description="Flexible Camera Calibration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SimonLBSoerensen/Flexible-Camera-Calibration",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="camera calibration",
    python_requires='>=3.6',
    license='MIT',
    install_requires=["numpy", "opencv-contrib-python>=3.1", "tqdm",
                      "matplotlib", "scipy", "Pillow", "seaborn", "pandas"

                      ],
)

#python setup.py sdist bdist_wheel
#python -m twine upload dist/*




import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d

import os
import flexiblecc as fcc

datasetpath = "../../CalImgs/cam_0*"


import glob
import numpy as np
import cv2
import matplotlib.pyplot as plt
from tqdm import tqdm


def axisEqual3D(ax):
    extents = np.array([getattr(ax, 'get_{}lim'.format(dim))() for dim in 'xyz'])
    sz = extents[:,1] - extents[:,0]
    centers = np.mean(extents, axis=1)
    maxsize = max(abs(sz))
    r = maxsize/2
    for ctr, dim in zip(centers, 'xyz'):
        getattr(ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)

'''
color_cal_imgs_files = glob.glob(datasetpath)
color_cal_imgs = np.array([cv2.imread(f) for f in tqdm(color_cal_imgs_files)])
gray_cal_imgs = [cv2.cvtColor(img, cv2.COLOR_BGR2GRAY) for img in tqdm(color_cal_imgs)]


pattern_size = (12, 12)


retval, cameraMatrix, distCoeffs, rvecs, tvecs, stdDeviationsIntrinsics, stdDeviationsExtrinsics, perViewErrors, points_3D, points_2D = fcc.OldSchoolCC.calibrate_camera(gray_cal_imgs, pattern_size, verbose=1)

np.savez("savedOldSchool", rvecs, tvecs, points_3D, points_2D)
'''

rvecs, tvecs, checkerboard_points, points_2D = dict(np.load("savedOldSchool.npz")).values()

checkerboard_points = checkerboard_points.reshape(checkerboard_points.shape[0], checkerboard_points.shape[2], checkerboard_points.shape[3], 1)

bad_images = np.array([4,5,6,7,8,17,18,19,20,21,22,23,35,36,37,38,40,41,42,43,44,53,54,55,56,61,62,63,64,68,73,74,75,81,82,83,84,85,93,94,39])
bad_images = bad_images - 1

rvecs = np.array(rvecs)
tvecs = np.array(tvecs)

import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d

ax = plt.axes(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')

cam1_points_3D = np.reshape(checkerboard_points[0], (-1,3))

ax.plot3D(cam1_points_3D[:,0], cam1_points_3D[:,1], cam1_points_3D[:,2], 'bo')
for i in range(len(checkerboard_points)):

    #if i in bad_images:
    #ax.plot3D(tvecs[i][0], tvecs[i][1], tvecs[i][2], 'ro')

    rot_mat = cv2.Rodrigues(rvecs[i])[0]
    p = np.dot(rot_mat, -tvecs[i])

    if(i == 38):
        ax.plot3D(p[0], p[1], p[2], 'gx')
    else:
        ax.plot3D(p[0], p[1], p[2], 'rx')


    #if(i == 38):
    #    ax.plot3D(-tvecs[i][0], -tvecs[i][1], -tvecs[i][2], 'gx')
    #else:
    #    ax.plot3D(-tvecs[i][0], -tvecs[i][1], -tvecs[i][2], 'rx')

plt.show()


n_images = 94
grid_width = 12
grid_height = 12
grid_size = grid_width * grid_height

grid = np.zeros((grid_size, 3))
meshgrid = np.transpose(np.meshgrid(np.arange(0,12),np.arange(0,12)))

meshgrid = np.reshape(meshgrid, (grid_size, 2))

grid[:, :-1] = meshgrid
grid = np.reshape(grid, grid.shape+(1,))

forward_projected_rays = np.ndarray((n_images, grid_size, 3, 1))
for i,(r,t) in enumerate(zip(rvecs, tvecs)):
    rotation_matrix = cv2.Rodrigues(r)[0]
    for j in range(grid_size):
        forward_projected_rays[i,j] = np.dot(rotation_matrix, checkerboard_points[i,j])+t



ax = plt.axes(projection='3d')
ax.set_xlabel('x')
ax.set_ylabel('y')
ax.set_zlabel('z')


ax.plot3D([0],[0],[0], 'rx')
#for i, c in enumerate(["b", "y", "g", "r", "c"]):
for i in range(len(forward_projected_rays)):

    #if i in bad_images:

    #draw normals
    vector1 = np.reshape(forward_projected_rays[i][12*5+5] - forward_projected_rays[i][12*5+6], (3,))
    vector2 = np.reshape(forward_projected_rays[i][12*5+5] - forward_projected_rays[i][12*6+5], (3,))
    normalvector = np.cross(vector1, vector2)
    pa = forward_projected_rays[i][12*5+5].reshape((3,))
    pb = pa - normalvector * 100
    ax.plot([pa[0], pb[0]], [pa[1], pb[1]],zs=[pa[2], pb[2]])

    #draw planes
    cam1_points_3D = np.reshape(forward_projected_rays[i], (-1,3))
    ax.plot_trisurf(cam1_points_3D[:,0], cam1_points_3D[:,1], cam1_points_3D[:,2])
    
axisEqual3D(ax)
plt.show()


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
from tqdm import tqdm

from central_model import CentralModel, fit_central_model

import pyqtgraph as pg
import pyqtgraph.opengl as gl
from OpenGL.GL import glReadBuffer, GL_FRONT

## GENERATION ##
fit = True

image_dimensions = (200, 200)
grid_dimensions = (200, 200)

order = 2
shape = (6,6,3)

knot_method = 'open_uniform' # 'uniform' or 'open_uniform'
 
end_divergence = 1e-10
min_basis_value = 0

## PLOTTING ##
export_option = 'single' # 'rotation', 'single' or 'none'. 
image_name = '{}_{}_{}'.format(order, fit, knot_method)

draw_as_surface = True

draw_pts = True
draw_ctrl = True
draw_tv = True

orbit = False

frame_time = 1000/60 # in ms

np.random.seed(2)
target_values = np.random.normal(0, np.sqrt(np.average(image_dimensions)), np.prod(shape))
target_values = np.reshape(target_values, shape)

if fit:
    cm, results = fit_central_model(
        target_values,
        image_dimensions=image_dimensions, 
        grid_dimensions=grid_dimensions, 
        order=order,
        knot_method=knot_method,
        end_divergence=end_divergence,
        min_basis_value=min_basis_value,
        verbose=2
    )
else:
    cm = CentralModel(
        image_dimensions,
        grid_dimensions,
        target_values,
        order=order,
        knot_method=knot_method,
        end_divergence=end_divergence,
        min_basis_value=min_basis_value
    )

d = np.divide(np.subtract(grid_dimensions, 1), np.subtract(shape[:-1], 1))
c = np.divide(np.subtract(grid_dimensions, image_dimensions), 2)

ctrl = np.round(np.array([[[i * d[0], j * d[1], cm.a[i,j,0]] for j in range(shape[1])] for i in range(shape[0])]))
ctrl = np.reshape(ctrl, (-1, 3))
ctrl[:,:2] -= c

tv = np.round(np.array([[[i * d[0], j * d[1], target_values[i,j,0]] for j in range(shape[1])] for i in range(shape[0])]))
tv = np.reshape(tv, (-1, 3))
tv[:,:2] -= c

pts = np.ndarray(grid_dimensions + (3,))

for u in tqdm(range(image_dimensions[0])):
    for v in range(image_dimensions[1]):
        pts[u, v, :] = cm.sample(u, v)

y = pts[:,:,0]

x = np.array([[[i, j, y[i,j]] for j in range(image_dimensions[1])] for i in range(image_dimensions[0])])
x = np.reshape(x, (-1,3))

# Make pyqtgraph window.
app = pg.mkQApp()
w = gl.GLViewWidget()
w.show()

## Adds point to scatter plot.
if draw_pts:
    ptcolor = np.ndarray((x.shape[0], 4))
    c = np.divide(np.subtract(x[:,2], np.min(x[:,2])), np.max(x[:,2]) - np.min(x[:,2]))
    ptcolor[:,0] = c
    ptcolor[:,1] = 0
    ptcolor[:,2] = np.subtract(1, c)
    ptcolor[:,3] = 1

    # Creates surface/pointcloud for the b-spline samples.
    if draw_as_surface:
        x1 = np.arange(np.max(x[:,0]) + 1)
        x2 = np.arange(np.max(x[:,1]) + 1)
        pts = gl.GLSurfacePlotItem(x1, x2, x[:,2].reshape(len(x1), len(x2)), colors=ptcolor)

    else:
        pts = gl.GLScatterPlotItem(pos=x, color=ptcolor, size=1)

    w.addItem(pts)

scatterPlotItems = {}

# Creates points for the control points of the b-spline.
if draw_ctrl:
    ctrlcolor = np.full((ctrl.shape[0], 4), np.array([0, 1, 0, 1]))
    scatterPlotItems['ctrl'] = gl.GLScatterPlotItem(pos=ctrl, color=ctrlcolor)
    w.addItem(scatterPlotItems['ctrl'])

# Creates points for the target points of the b-spline.
if draw_tv:
    tvcolor = np.full((ctrl.shape[0], 4), np.array([1, 1, 1, 1]))
    scatterPlotItems['tv'] = gl.GLScatterPlotItem(pos=tv, color=tvcolor)
    w.addItem(scatterPlotItems['tv'])

# Orients camera in 3d.
w.showFullScreen()
centerpoint = np.average(x, axis=0)
w.pan(centerpoint[0], centerpoint[1], centerpoint[2])
w.setCameraPosition(distance=400)


# Orbiting and image exporting.
da = 0.2
current_angle = 0
i = 0
def update():
    global current_angle, da, i, export_option

    if export_option == 'single':
        w.grabFrameBuffer().save('images/{}.png'.format(image_name))
        export_option = 'none'

    if current_angle < 720 and export_option == 'rotation':
        w.grabFrameBuffer().save('images/{}_{}.png'.format(image_name, i))
        current_angle += da

    if orbit:
        w.orbit(da, 0)
        
    i += 1

timera = pg.QtCore.QTimer()
timera.timeout.connect(update)

if export_option == 'rotation':
    timera.start(1)
else:
    timera.start(frame_time)

glReadBuffer(GL_FRONT) # A little bit of c-code in my lines

# Executes application
app.exec_()


import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import os
import uuid
from PIL import Image
import glob
import shutil


def make_cm_vec_plot(cm, image_dimensions_step):
    w, h = (cm.image_width, cm.image_height)

    w_cor = image_dimensions_step if w % image_dimensions_step == 0 else 0
    h_cor = image_dimensions_step if h % image_dimensions_step == 0 else 0

    points = np.mgrid[0:w + w_cor:image_dimensions_step, 0:h + h_cor:image_dimensions_step].T.reshape(-1, 2)

    sample_vec = lambda el: cm.sample(el[0], el[1])

    vecs = np.array(list(map(sample_vec, points)))

    X, Y = points.T
    Z = np.zeros_like(X)

    U, V, W = vecs.T

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.quiver(X, Y, Z, U, V, W, arrow_length_ratio=0)

    ax.set_xlim(0, w)
    ax.set_ylim(0, h)
    ax.set_zlim(0, np.max(W) * 2)
    plt.tight_layout()

    return ax


def make_cm_vec_plot_gif(cm, image_dimensions_step, gif_file_out, azim_start=0, azim_end=360, azim_step=5, elev_start=0,
                         elev_end=90, elev_step=10, duration=150, loop=0, dpi=None, show=False):
    elev_cor = elev_step if elev_step != 360 else 0
    azim_cor = azim_step if azim_step != 360 else 0

    img_idx = 0
    runId = str(uuid.uuid4())
    os.makedirs(runId)

    ax = make_cm_vec_plot(cm, image_dimensions_step)

    for elev in range(elev_start, elev_end + elev_cor, elev_step):
        for azim in range(azim_start, azim_end + azim_cor, azim_step):
            ax.view_init(elev, azim)
            plt.draw()

            plt.savefig(os.path.join(runId, "image_{}.jpg".format(img_idx)), dpi=dpi)
            img_idx += 1

            if show:
                plt.pause(.001)
    if not show:
        plt.close()

    fp_in = os.path.join(runId, "image_*.jpg")
    fp_out = gif_file_out

    files = np.array(sorted(glob.glob(fp_in)))
    sort_idxs = np.argsort([int(f.split(".")[0].split("_")[-1]) for f in files])
    files = files[sort_idxs]

    img, *imgs = [Image.open(f) for f in files]
    img.save(fp=fp_out, format='GIF', append_images=imgs,
             save_all=True, duration=duration, loop=loop)

    shutil.rmtree(runId)
